import os
from tensorflow import gfile
import shutil
# from pyspark import SparkFiles
import smart_open
try:
    from dl_rank.file.file_api import s3FileObj, pathStr,hdfsFile,s3File,addtail
except:
    import file_api
    s3FileObj = file_api.s3FileObj
    pathStr = file_api.pathStr
    hdfsFile = file_api.hdfsFile
    s3File = file_api.s3File
    from .file_api import addtail

def typeDiagnosis(*path_vars):
    def file_func(func):
        func_arg_vars = list(func.__code__.co_varnames[:func.__code__.co_argcount])
        path_like_list = path_vars if len(path_vars) else func_arg_vars
        def wrapper(*args, **kwargs):
            # func_var_dict = {var: kwargs.pop(var) if var in kwargs else args.pop(0) for var in func_arg_vars if var in kwargs or len(args)>0}
            # func_var_dict = {var_name: pathStr(var_val).path if var_name in path_like_list else var_val
            #                  for var_name, var_val in func_var_dict.items()}

            args_name = [var for var in func_arg_vars if var not in kwargs][:len(args)]
            ftype = [pathStr(kwargs[path_var]).attr if path_var in kwargs else pathStr(args[args_name.index(path_var)]).attr
                     for path_var in path_like_list]
            args = [pathStr(arg).path if name in path_like_list else arg for name, arg in zip(args_name, args)]
            kwargs = {k: pathStr(arg).path if arg in path_like_list else arg for k, arg in kwargs.items()}

            if all([ftype[0] == ftype_ for ftype_ in ftype]):
                File = wfile.filesys.get(ftype[0], gfile)
                if hasattr(File, func.__name__):
                    filefunc = getattr(File, func.__name__)
                elif hasattr(gfile, func.__name__):
                    filefunc = getattr(gfile, func.__name__)
                else:
                    filefunc = func
                return filefunc(*args, **kwargs)
            else:
                return func(*args, **kwargs)
        return wrapper
    return file_func


class wfile(object):
    path_store = dict()
    filesys = {'s3': s3File, 'hdfs':hdfsFile, 'local': gfile}
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, '_instance'):
            cls._instance = super().__new__(cls, *args, **kwargs)
        return cls._instance

    @staticmethod
    def init_hdfs(ip_port=''):
        if ip_port == '':
            import socket
            ip_port = socket.gethostname() + '.us-west-2.compute.internal' + ':50070'
        hdfsFile.init_client(ip_port)

    @staticmethod
    @typeDiagnosis()
    def Exists(filename):
        pass

    @staticmethod
    def _overwrite(objpath, is_dir):
        ftype = wfile.filesys.get(pathStr(objpath).attr, gfile)
        exist = ftype.Exists(objpath)
        if exist:
            if is_dir:
                ftype.DeleteRecursively(objpath)
            else:
                ftype.Remove(objpath)

    @staticmethod
    def Open(*args, **kwargs):
        args = [arg.path if isinstance(arg, pathStr) else arg for arg in args]
        for key, value in kwargs.items():
            if isinstance(value, pathStr):
                kwargs[key] = value.path
        return smart_open.open(*args, **kwargs)

    @staticmethod
    @typeDiagnosis()
    def DeleteRecursively(dirname):
        pass

    @staticmethod
    @typeDiagnosis()
    def IsDirectory(dirname):
        pass

    @staticmethod
    @typeDiagnosis()
    def Walk(top):
        pass

    @staticmethod
    @typeDiagnosis()
    def MakeDirs(dirname):
        pass

    @staticmethod
    @typeDiagnosis()
    def Remove(filename):
        pass

    @staticmethod
    @typeDiagnosis()
    def ListDirectory(dirname):
        pass

    @staticmethod
    @typeDiagnosis()
    def Rename(oldname, newname):
        pass

    @staticmethod
    @typeDiagnosis('srcPath', 'dstPath')
    def RenameDirectory(srcPath, dstPath):
        path1, path2 = addtail(srcPath), addtail(dstPath)
        if path1 != path2:
            wfile._CopyDirectory(path1, path2)
            wfile.DeleteRecursively(path1)

    @staticmethod
    def _CopyDirFuncDiffSys(*ftypelist):
        if 's3' in ftypelist and 'hdfs' in ftypelist:
            copy_func = s3File.CopyDirHdfs
        elif 's3' in ftypelist and 'local' in ftypelist:
            copy_func = s3File.CopyDirLocal
        elif all(['local' == ftype for ftype in ftypelist]):
            copy_func = shutil.copytree
        else:
            copy_func = wfile._CopyDirectory
        return copy_func

    @staticmethod
    def _CopyFuncDiffSys(*ftypelist):
        # if 's3' in ftypelist and 'hdfs' in ftypelist:
        #     copy_func = s3File.CopyHdfs
        if 's3' in ftypelist and 'local' in ftypelist:
            copy_func = s3File.CopyLocal
        elif 'hdfs' in ftypelist and 'local' in ftypelist:
            copy_func = hdfsFile.CopyLocal
        elif 'hdfs' in ftypelist and 's3' in ftypelist:
            copy_func = hdfsFile.CopyS3
        else:
            copy_func = gfile.Copy
        return copy_func

    @staticmethod
    def _CopyDirectory(srcPath, dstPath):
        for root, dirs, files in wfile.Walk(srcPath):
            new_root = dstPath + root[len(srcPath):]
            for d in dirs:
                wfile.MakeDirs(os.path.join(new_root, d))
            for f in files:
                wfile.Copy(os.path.join(root, f), os.path.join(new_root, f))

    @staticmethod
    @typeDiagnosis('srcPath', 'dstPath')
    def CopyDirectory(srcPath, dstPath, overwrite=False):
        file1, file2 = pathStr(srcPath).attr, pathStr(dstPath).attr
        path1, path2 = addtail(srcPath), addtail(dstPath)
        if path1 != path2:
            if overwrite:
                wfile._overwrite(path2, is_dir=True)
            copy_func = wfile._CopyDirFuncDiffSys(file1, file2)
            copy_func(path1, path2)

    @staticmethod
    @typeDiagnosis('srcPath', 'dstPath')
    def Copy(srcPath, dstPath, overwrite=False):
        # src and dst path in different file system
        file1, file2 = pathStr(srcPath).attr, pathStr(dstPath).attr
        path1, path2 = srcPath, dstPath
        if path1 != path2:
            if overwrite:
                wfile._overwrite(path2, is_dir=False)
            try:
                gfile.Copy(path1, path2)
            except:
                copy_func = wfile._CopyFuncDiffSys(file1, file2)
                copy_func(srcPath, dstPath)

if __name__ == '__main__':
    a = pathStr('/Users/snoworday/Downloads/part-00100-eb6b4732-3b25-4cd2-978b-7f4f1f5f18a5-c000.txt')
    f = wfile.Open(a, 'r')
    c = 1


from __future__ import absolute_import
import paramiko
import os
import json
import inspect
import subprocess
import argparse
from dl_rank.file.fileUtils import *
from dl_rank.file import wfile, pathStr
import time
try:
    from dl_rank.utils.modelUtils import setEnv
except:
    from utils.modelUtils import setEnv
from multiprocessing import Pool

# names about EMR
UserName = 'hadoop'
port = 22
tf_port = 2222
emr_home_path = '/home/hadoop'

_project_path = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
_exec_file = 'solo.py'


class SFTPClient(paramiko.SFTPClient):
    def put_dir(self, source, target):
        ''' Uploads the contents of the source directory to the target path. The
            target directory needs to exists. All subdirectories in source are
            created under target.
        '''

        for item in os.listdir(source):
            if os.path.isfile(os.path.join(source, item)):
                try:
                    self.put(os.path.join(source, item), '%s/%s' % (target, item))
                except:
                    try:
                        self.remove('%s/%s' % (target, item))
                        self.put(os.path.join(source, item), '%s/%s' % (target, item))
                    except:
                        pass
            else:
                self.mkdir('%s/%s' % (target, item), ignore_existing=True)
                self.put_dir(os.path.join(source, item), '%s/%s' % (target, item))

    def mkdir(self, path, mode=511, ignore_existing=True):
        ''' Augments mkdir by adding an option to not fail if the folder exists  '''
        try:
            super(SFTPClient, self).mkdir(path, mode)
        except IOError:
            if ignore_existing:
                pass
            else:
                raise

def getClusterId(emrName):
    cmd = r'''
    aws emr list-clusters  --active | jq '.Clusters' | jq '.[]' -c | grep  {} | jq '.Id' | tr -d '"'
    '''.format(emrName)
    cmd_out = os.popen(cmd).read()
    clusterId = cmd_out.split('\n')[0]
    return clusterId

def getHostnameList_Chief_Hostnum(emrName):
    cmd = r'''
    aws emr list-instances --cluster-id {} | jq '.Instances' | jq '.[]' -c | jq '.PrivateIpAddress' | tr -d '"'
    '''.format(getClusterId(emrName))
    cmd_out = os.popen(cmd).read()
    hostnameList = cmd_out.split('\n')[:-1]

    cmd = r'''
    aws emr list-instances --cluster-id {} | grep 'PrivateIpAddress' | wc -l
    '''.format(getClusterId(emrName))
    cmd_out = os.popen(cmd).read()
    hostnum = cmd_out.split('\n')[0]

    cmd = r'''
    ifconfig | grep 'inet addr:172' | awk '{print $2}' | sed 's/addr://g'
    '''
    cmd_out = os.popen(cmd).read()
    chief = cmd_out.split('\n')[0]
    return hostnameList, chief, hostnum

def generateTFConfig(worker_ps_evaluator, chief, ps_num, ev_num):
    def addPort(nodelist, port=tf_port):
        out = [n+':'+str(port) for n in nodelist]
        return out
    assert ps_num + ev_num <= len(worker_ps_evaluator), 'no sufficient node remaining'
    if ev_num == 0:
        worker, ps = worker_ps_evaluator[:-(ps_num + ev_num)], worker_ps_evaluator[-(
                    ps_num + ev_num):]
        evaluator = []
    else:
        worker, ps, evaluator = worker_ps_evaluator[:-(ps_num+ev_num)], worker_ps_evaluator[-(ps_num+ev_num):-ev_num], worker_ps_evaluator[-ev_num:]

    TF_CONFIG_dict = dict()
    cluster = {'chief': addPort([chief]), 'worker': addPort(worker)}
    if ps_num:
        cluster.update({'ps': addPort(ps)})
    for type, nodes in cluster.items():
        idx = 0
        for node in nodes:
            task = {'type': type, 'index': idx}
            TF_CONFIG_dict[node.split(':')[0]] = {'cluster': cluster, 'task': task}
            idx += 1
    idx = 0
    for node in evaluator:
        task = {'type': 'evaluator', 'index': idx}
        TF_CONFIG_dict[node] = {'cluster': cluster, 'task': task}
        idx += 1
    return TF_CONFIG_dict

def transportFile(ssh, file_path_list, target_path):
    transport = ssh.get_transport()
    sftp = SFTPClient.from_transport(transport)
    try:
        sftp.rmdir(target_path)
    except:
        pass
    sftp.mkdir(target_path)
    for file_path in file_path_list:
        target_path_ = os.path.join(target_path, file_path.rsplit('/', 1)[1])
        if not wfile.IsDirectory(file_path):
            sftp.put(file_path, target_path_)
        else:
            sftp.mkdir(target_path_)
            sftp.put_dir(file_path, target_path_)

def distributeMulti(mode, *args, **kwargs):
    func = distributeTrain if mode == 'train' else distributeInfer
    p = Pool()
    result = p.apply(func, args, kwargs)
    p.close()
    p.join()
    pid = result
    return pid


def distributeTrain(confDir, date, retrain, ps_num, ev_num, emrName, keyFile, logDir=emr_home_path, hdfspath=None, dirs2issue=None):
    if not wfile.Exists(logDir):
        wfile.MakeDirs(logDir)
    confDir = pathStr(confDir)

    if retrain:
        clear_summary(confDir)
    if dirs2issue is None:
        dirs2issue = []
    hostnameList, chief, hostnum = getHostnameList_Chief_Hostnum(emrName)
    hostnameList.pop(hostnameList.index(chief))
    TF_CONFIG_dict = generateTFConfig(hostnameList, chief, ps_num, ev_num)
    total_worker = int(hostnum) - ev_num - ps_num
    worker_wi, ps_wi = 0, 0
    date = '--date {}'.format(date) if date!='' else ''
    dirs2issue.append(confDir)
    cmd_dir = os.path.join(logDir, 'cmd')
    if not wfile.Exists(cmd_dir):
        wfile.MakeDirs(cmd_dir)
    def run_remote(wi, TF_CONFIG, UserName, hostname, is_eval):
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh.connect(hostname=hostname, port=port, username=UserName, key_filename=keyFile)
        if len(dirs2issue) > 0:
            transportFile(ssh, dirs2issue, logDir)
        if __name__=='__main__':
            trainpath = os.path.join(logDir, dirs2issue[0].split('/')[-1], _exec_file)
        else:
            trainpath = '-m dl_rank.solo'
        cmd = r'''
        python3 {trainpath} --mode train --tw {tw} --wi {wi} --ps {ps} {date} --conf {conf} --tfconfig {tfconfig} --logpath {logpath} --hdfspath {hdfs}
        '''.format(trainpath=trainpath, tw=total_worker if not is_eval else 1, wi=wi, ps=ps_num,
                   date=date, conf=confDir, tfconfig=TF_CONFIG, logpath=logDir, hdfs=hdfspath)
        cmd = cmd.strip() + ' > {}'.format(os.path.join(logDir, 'train.log'))
        with wfile.Open(cmd_dir+'/{}'.format(hostname), 'w+') as f:
            f.writelines(cmd)
        # subprocess.Popen(['ssh', ' -i {} '.format(keyFile), UserName + '@' + hostname, nohup_cmd], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        _ = ssh.exec_command(cmd)
        time.sleep(0.1)
        ssh.close()

    def run_local(wi, TF_CONFIG):
        if __name__ == '__main__':
            trainpath = os.path.join(dirs2issue[0], _exec_file)
        else:
            trainpath = '-m dl_rank.solo'
        cmd = '''
        python3 {trainpath} --mode train --tw {tw} --wi {wi} --ps {ps} {date} --conf {conf} --tfconfig {tfconfig} --logpath {logpath} --hdfspath {hdfs}
        '''.format(trainpath=trainpath, tw=total_worker, wi=wi, ps=ps_num, date=date,
                   conf=confDir, tfconfig=TF_CONFIG, logpath=logDir, hdfs=hdfspath)
        # nohup_cmd = '/home/hadoop/nohup' + cmd + ' > {} 2>&1 &'.format(os.path.join(logDir, 'dl_rank_run.log'))
        cmd = cmd.strip() + ' > {}'.format(os.path.join(logDir, 'train.log'))
        with wfile.Open(cmd_dir+'/local', 'w+') as f:
            f.writelines(cmd)
        process = subprocess.Popen([cmd], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
        return process

    for hostname, TF_CONFIG in TF_CONFIG_dict.items():
        node_type = TF_CONFIG['task']['type']
        if node_type in ['chief', 'worker']:
            wi = worker_wi
            worker_wi += 1
        elif node_type == 'ps':
            wi = ps_wi
            ps_wi += 1
        else:
            wi = 0
        TF_CONFIG = json.dumps(TF_CONFIG).replace(' ', '').replace('"', '_').replace(',', '*')
        if node_type == 'chief':
            chief = run_local(wi, TF_CONFIG)
        else:
            run_remote(wi, TF_CONFIG, UserName, hostname, is_eval=node_type == 'evaluator')
    return chief.pid

def stopCluster(emrName, keyFile):
    hostnameList, chief, _ = getHostnameList_Chief_Hostnum(emrName)
    for node in hostnameList:
        if node == chief:
            os.system("ps -ef | grep python | grep solo | awk '{print \"kill -9 \" $2}' | bash -v")
        else:
            ssh = paramiko.SSHClient()
            ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            ssh.connect(hostname=node, port=port, username=UserName, key_filename=keyFile)
            _ = ssh.exec_command("ps -ef | grep python | grep solo | awk '{print \"kill -9 \" $2}' | bash -v")
            time.sleep(0.1)
            ssh.close()

def _zipDir(dirList, project_path):
    from zipfile import ZipFile, ZIP_DEFLATED
    py_file = []
    for dir in dirList:
        dir_path = os.path.join(project_path, '{}'.format(dir))
        if os.path.exists(dir_path + '.zip'):
            os.remove(dir_path + '.zip')
        zipfile = ZipFile(dir_path+'.zip', 'w', ZIP_DEFLATED)
        for file in os.listdir(dir_path):
            zipfile.write(os.path.join(dir_path, file))
        zipfile.close()
        py_file.append(dir_path + '.zip')
    return py_file


def _addConfig(conf_path):
    if conf_path.startswith('s3'):
        return []
    py_files = []
    for file in wfile.IsDirectory(conf_path):
        py_files.append(os.path.join(conf_path, file))
    assert py_files != [], 'Cant find conf dir'
    return py_files

def clear_summary(conf_path):
    import yaml
    with wfile.Open(conf_path + '/mission.yaml', 'r') as f:
        train_conf = yaml.load(f)
        model_dir = train_conf['train']['model_dir']
    if wfile.Exists(model_dir):
        wfile.DeleteRecursively(model_dir)

def run_tensorboard(conf='', model_dir='', logpath=emr_home_path, port=6006):
    assert conf!='' or model_dir!='', 'Please define summary path'
    if model_dir != '':
        pass
    else:
        import yaml
        conf_path = copy_conf2logpath(conf, logpath)
        with wfile.Open(conf_path + '/mission.yaml', 'r') as f:
            train_conf = yaml.load(f)
            model_dir = train_conf['train']['model_dir']

    nohup_cmd = '/usr/bin/nohup python3 -m tensorboard.main --logdir={model_dir} --port={port} > {logpath} 2>&1 &'.format(
        model_dir=model_dir, port=port, logpath=os.path.join(logpath, 'dl_rank_tensorboard.log'))
    process = subprocess.Popen([nohup_cmd], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    return process

def distributeInfer(conf, date, ps, emrName, logpath=emr_home_path, hdfspath=None, use_TFoS=True, dirs2issue=None, executer_num=0):
    _args_useSpark = True
    if dirs2issue is None:
        dirs2issue = []
    conf = pathStr(conf)
    if __name__=='__main__':
        py_file_conf = _addConfig(conf)
        package_names = ['model', 'utils', 'conf', 'layer', 'file']
        py_file_zip = _zipDir(package_names, _project_path)
        py_file = ','.join(py_file_zip + py_file_conf + dirs2issue)
    else:
        py_file_conf = _addConfig(conf)
        py_file = ','.join(py_file_conf + dirs2issue)

    py_file = '--py-files {}'.format(py_file) if py_file != '' else ''
    setEnv()
    if executer_num==0:
        _, _, executer_num = getHostnameList_Chief_Hostnum(emrName)
        if use_TFoS:
            executer_num = int(executer_num) - 1

    use_TFoS_param = '--executor-cores 1' if use_TFoS else ''
    date = '--date {}'.format(date) if date != '' else ''
    command = r'''
    spark-submit --master yarn {py_file} {archive} {TFoS} --num-executors {executor} --executor-memory 15G --driver-memory 15G --conf spark.executorEnv.JAVA_HOME="$JAVA_HOME" --conf spark.executorEnv.PYSPARK_PYTHON="/usr/bin/python3" --conf spark.executorEnv.PYSPARK_DRIVER_PYTHON="/usr/bin/python3"  --conf spark.executorEnv.CLASSPATH="$($HADOOP_HOME/bin/hadoop classpath --glob):${{CLASSPATH}}"  --conf spark.executorEnv.LD_LIBRARY_PATH="/usr/lib/hadoop/lib/native:${{JAVA_HOME}}/lib/amd64/server" {exec_file} {date} --mode infer --useSpark --logpath {tf_log_path} --hdfspath {hdfs} --ps {ps} --num_executor {num_executor} {use_TFoS} --conf {conf} > {inferlogpath} 
    '''.format(py_file=py_file, archive= '', TFoS=use_TFoS_param, date=date, executor=executer_num, exec_file=os.path.join(_project_path, _exec_file),
               tf_log_path=logpath, hdfs=hdfspath, ps=ps, num_executor=executer_num, use_TFoS='--use_TFoS' if use_TFoS else '', conf=conf,
               inferlogpath=os.path.join(logpath, 'infer.log')
               )
    process = subprocess.Popen([command], shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    return process.pid

if __name__=='__main__':
    '''
    python3 symphony.py --conf I2Iconf_uv --date 2019-07-24:2019-08-04 --mode train --ps 1 --ev 1
    '''
    parser = argparse.ArgumentParser()
    parser.add_argument('--ps', type=int, help='number of parameter server', default=1)
    parser.add_argument('--ev', type=int, help='number of evaluator', default=1)
    parser.add_argument('--emr', type=str, help='name of cluster', default='wangqi')
    parser.add_argument('--keyFile', type=str, help='keyfile path on chief node', default='/home/hadoop/wangqi.pem')
    parser.add_argument('--date', type=str, help='date of dataset', default='')
    parser.add_argument('--conf', type=str, help='select a config in config dir', default='')
    parser.add_argument('--logpath', type=str, help='log dir path', default=emr_home_path)
    parser.add_argument('--mode', type=str, help='train|infer|stop|export|tb', default='train')
    parser.add_argument('--port', type=int, help='default port for tensorboard', default=6006)
    parser.add_argument('--useSpark', help='use spark-file when infer', action='store_true')
    parser.add_argument('--retrain', help='if retrain clear summary', action='store_true')
    parser.add_argument('--use_TFoS', help='use tensorflowonspark when infer', action='store_true')

    _args = parser.parse_args()
    # sth static
    EmrName = _args.emr
    key_filepath = _args.keyFile
    # config of cluster
    ps_num = _args.ps
    ev_num = _args.ev
    _args.logpath = os.path.join(os.getcwd(), _args.logpath)
    if not gfile.Exists(_args.logpath):
        gfile.MakeDirs(_args.logpath)

    if _args.mode == 'train':
        assert _args.conf != '', 'please set conf!'
        distributeTrain(_args.conf, _args.date, _args.retrain,
                        ps_num, ev_num, EmrName, _args.keyFile, _args.logpath, [_project_path])
    elif _args.mode == 'stop':
        stopCluster(EmrName, _args.keyFile)
    elif _args.mode == 'infer':
        distributeInfer(_args.conf, _args.date, ps_num, EmrName, _args.logpath, _args.use_TFoS)
    elif _args.mode == 'tb':
        run_tensorboard(conf=_args.conf, port=_args.port, logpath=_args.logpath)
    elif _args.mode == 'export':
        os.system("python3 {exec_file} --conf {conf} --mode export".format(exec_file=os.path.join(_project_path, _exec_file)
                                                                         , conf=_args.conf))
    elif _args.mode == 'export_online':
        os.system("python3 {exec_file} --conf {conf} --mode export_online".format(exec_file=os.path.join(_project_path, _exec_file)
                                                                         , conf=_args.conf))





from yaml import SafeDumper
import yaml
from tensorflow import gfile
import os

SafeDumper.add_representer(
    type(None),
    lambda dumper, value: dumper.represent_scalar(u'tag:yaml.org,2002:null', '')
  )

class ConfigManager(object):
    def __init__(self, config_dict=None, **template):
        self._yamls = dict()
        if config_dict is None:
            self.inputConfig(template)
        elif template is None:
            self.inputConfig(config_dict)
        else:
            config_dict.update(template)
            self.inputConfig(config_dict)

    def dumpConfig(self, save_path):
        for yamlName in self._yamls:
            out_path = os.path.join(save_path, yamlName+'.yaml')
            with gfile.GFile(out_path, 'w') as f:
                yaml.safe_dump(self._yamls[yamlName], f, default_flow_style=False)

    def getConfig(self):
        return self._yamls

    def inputConfig(self, template):
        for yamlName, value in template.items():
            self.addYaml(yamlName, value)

    def addYaml(self, yamlName, value):
        if isinstance(value, str):
            with gfile.GFile(value, 'r') as f:
                self._yamls[yamlName] = yaml.load(f)
        elif isinstance(value, dict):
            self._yamls[yamlName] = value

    def update(self, input_dict):
        # for each yaml
        for yaml_name, config_dict in input_dict.items():
            # for each {(key1, key2, key3...): value} ...
            for multi_key, value in config_dict.items():
                multi_dict = self._yamls[yaml_name]
                if isinstance(multi_key, str):
                    multi_dict[multi_key] = value
                else:
                    for idx in range(len(multi_key)-1):
                        multi_dict = multi_dict[multi_key[idx]]
                    multi_dict[multi_key[-1]] = value

    def check(self):
        pass



if __name__ == '__main__':
    conf = ConfigManager(f=
                       {'type': 'continuous', 'transform': None, 'parameter':
                           {'normalization': None, 'boundaries': None},
                        'ignore': False, 'scope': ['wide', 'deep']})
    conf.update({'f': {('parameter', 'normalization'): 3, ('transform', ): 5}})
    a = 1




from dl_rank.file.file_api import pathStr
from dl_rank.file.wfile import wfile
import os
from tensorflow import gfile

def copy_conf2logpath(conf_dir, save_path):
    conf_dir = pathStr(conf_dir)
    conf_path, conf_name = conf_dir.rsplit('/', 1)
    out_path = pathStr(conf_name, prefix=save_path)
    if not wfile.Exists(out_path):
        wfile.MakeDirs(out_path)
    wfile.CopyDirectory(conf_dir, out_path, overwrite=True)
    return out_path



def getConfInfo(confDir):
    if confDir[-1] == '/':
        confDir = confDir[:-1]
    if confDir.startswith('s3'):
        return confDir, confDir
    else:
        confDir = os.path.abspath(confDir)
        confName = confDir.rsplit('/', 1)[1]
    return confName, confDir



def find_conf(conf, save_path):
    if conf.startswith('s3'):
        conf = conf.strip('/')
        conf_name = conf.rsplit('/', 1)[1]
        save_path = os.path.join(save_path, conf_name) + '/'
        gfile.MakeDirs(save_path)
        check = os.system('aws s3 cp {} {} --recursive'.format(conf+'/', save_path))
        conf = save_path
        if check:
            assert False, 'cant find conf in: {}'.format(conf)
        return conf
    conf_path = os.path.abspath(conf)
    return conf_path


from .BaseModel import baseModel
from .wide_and_deep import wdl
from .xcross import xcross
from .xdeepfm import xdeepfm
from .dcn import dcn
from .deepfm_v2 import deepfm as deepfm_v2
from .deepfm import deepfm
from .wide_deep_traditional_attention import wide_and_deep as wide_and_deep_traditional_attention

import os
import importlib.util


class modelFactory(object):
    model_dict = {'wdl': wdl, 'xcross': xcross, 'dcn': dcn, 'xdeepfm': xdeepfm, 'deepfm': deepfm,
                  'deepfm_v2': deepfm_v2, 'wide_and_deep_traditional_attention': wide_and_deep_traditional_attention}
    external_path = None
    @staticmethod
    def build(train_conf, model_conf, mode):
        model_type = train_conf['model_type']
        if model_type not in modelFactory.model_dict:
            exter_module_path = os.path.join(modelFactory.external_path, model_type+'.py')
            if os.path.exists(exter_module_path):
                exter_spec = importlib.util.spec_from_file_location('dl_rank.model.'+model_type, exter_module_path)
                exter_module = importlib.util.module_from_spec(exter_spec)
                exter_spec.loader.exec_module(exter_module)
                modelFactory._find_submodel(exter_module)
            else:
                assert False, 'Cant find external model: {}'.format(exter_module_path)
        model = modelFactory.model_dict[model_type](model_conf, mode)
        return model

    @staticmethod
    def register(model):
        name = model.name if hasattr(model, 'name') else model.__name__
        if name in modelFactory.model_dict:
            print('name: {} has been used'.format(name))
        else:
            modelFactory.model_dict.update({name: model})

    @staticmethod
    def _find_submodel(module):
        for attr in dir(module):
            attrObj = getattr(module, attr)
            try:
                if not isinstance(attrObj, baseModel) and issubclass(attrObj, baseModel):
                    modelFactory.register(attrObj)
            except:
                pass




from abc import ABCMeta, abstractmethod
from collections import OrderedDict, defaultdict
from functools import reduce
import numpy as np
import warnings
import tensorflow as tf
if type(tf.contrib) != type(tf): tf.contrib._warning = None
if tf.__version__[0]=='1':
    lookup = tf.contrib.lookup
else:
    lookup = tf.lookup
    assert False, 'emm.., lookup.index_table_from_tensor has been removed in TF2 ... and no substitute Q_Q'

class baseModel(metaclass=ABCMeta):
    name = 'base'
    def __init__(self, model_conf, mode):
        # self.activation_fn = activation_fn
        # self.initializer_fn = initializer_fn
        self.model_conf = model_conf
        self.mode = mode
        self.num_shards = self.model_conf['num_shards'] if 'num_shards' in self.model_conf else 1
        self.placeholder_map = self.model_conf['input_node_map'] if 'input_node_map' in self.model_conf else dict()
        self.out_node_names = self.model_conf['out_node_names'] if 'out_node_names' in self.model_conf else ['out']

    def set_embedding_parser(self, fn):
        self.embedding_parser_fn = fn(self._secondary_parse_fn)

    @staticmethod
    def build_embedding(params, num_shards):
        feature_conf = params['feature_conf']
        feature_list_conf = params['feature_list']
        feature_list = [feature_list_conf[key] for key in sorted(feature_list_conf, reverse=False)]
        model_conf = params['model_conf']
        vocabulary_conf = params['vocabulary_conf']
        embed_dim = model_conf['embed_dim']
        first_order = int(model_conf['first_order'])
        partitioner = tf.fixed_size_partitioner(num_shards) if num_shards > 1 else None

        table = OrderedDict()
        sparse = []
        deep = OrderedDict()
        multi = OrderedDict()
        model_struct = defaultdict(list)
        numeric = []
        dense = []
        dense_tag = []
        wide_dim, deep_dim, deep_num, cate_num, con_num, all_num, con_deep_num, con_bias_num = 0, 0, 0, 0, 0, 0, 0, 0
        for feature in feature_list:
            if not feature in feature_conf:
                continue
            conf = feature_conf[feature]
            if conf['ignore']:
                continue
            f_type, f_tran, f_param = conf['type'], conf['transform'], conf['parameter']
            if 'group' in conf:
                for struct in conf['group']:
                    model_struct[struct].append(feature)
            f_multi = conf['multi'] if 'multi' in conf else {'num': 1, 'same': True, 'combiner': 'none'}
            feature_name = f_param['name'] if 'name' in f_param else feature
            feature_embed_dim = f_param['embed_dim'] if 'embed_dim' in f_param else embed_dim
            feature_scope = f_param['scope'] if 'scope' in f_param else 'embedding'
            feature_initializer = conf['initializer'] if 'initializer' in conf else dict()

            with tf.variable_scope(feature_scope, reuse=tf.AUTO_REUSE, partitioner=partitioner) as scope:
                if f_type == 'category':
                    f_num, combiner = f_multi['num'], f_multi['combiner']
                    if combiner != 'none' and f_num >= 1:
                        f_num = 1
                    if f_tran == 'vocabulary_list':
                        vocabulary = vocabulary_conf[feature]
                        vocabulary = vocabulary ##########['DEFAULT'] + vocabulary
                        default_value = 0 if 'default' not in f_param \
                            else vocabulary.index(f_param['default']) \
                            if isinstance(f_param['default'], str) else f_param['default']
                        tail_value = 0 if 'tail' not in f_param \
                            else vocabulary.index(f_param['tail']) \
                            if isinstance(f_param['tail'], str) else f_param['tail']
                        fill_value = '' if 'fill' not in f_param \
                            else vocabulary[f_param['fill']] \
                            if isinstance(f_param['fill'], int) else f_param['fill']

                        table.update(
                            {feature: lookup.index_table_from_tensor(mapping=tf.constant(vocabulary), default_value=default_value)})

                        f_dim = len(vocabulary)*f_num
                        f_size = len(vocabulary)

                    elif f_tran == 'vocabulary_file':
                        default_value = 0 if 'default' not in f_param else f_param['default']
                        tail_value = 0 if 'tail' not in f_param else f_param['tail']
                        fill_value = '' if 'fill' not in f_param else f_param['fill']

                        table.update(
                            {feature: lookup.index_table_from_file(
                                vocabulary_file=file_path, vocab_size=vocab_size, default_value=default_value,delimiter='\t')
                            })
                        f_dim = vocab_size * f_num
                        f_size = vocab_size

                    elif f_tran == 'tabled':
                        f_dim = f_param['size']*f_num
                        f_size = f_param['size']
                        fill_value = f_param['fill'] if 'fill' in f_param else ''
                        default_value = f_param['default'] if 'default' in f_param else 0
                        tail_value = f_param['tail'] if 'tail' in f_param else 0
                    else:
                        assert False, 'only support category features with vocabulary or tabled'

                    if 'onehot' in conf['style']:
                        sparse.append(feature)
                        if f_num >= 1:
                            wide_dim += f_dim * f_num
                        else:
                            wide_dim += f_dim * (-f_num)
                    if 'embedding' in conf['style']:
                        deep.update({feature: tf.get_variable(initializer=get_initializer(**feature_initializer),
                                                              shape=[f_size, feature_embed_dim+first_order],
                                                              name='{}_embedding'.format(feature_name))})

                        if f_num >= 1:
                            deep_dim += (feature_embed_dim+first_order) * f_num
                            deep_num += f_num
                        else:
                            deep_dim += (feature_embed_dim + first_order) * (-f_num)
                            deep_num += -f_num

                    f_multi['same'] = False
                    dense_tag += [0] * abs(f_num)
                    cate_num += abs(f_num)
                    all_num += abs(f_num)

                elif f_type == 'numeric':
                    f_size = 1
                    f_num = f_multi['num']
                    numeric.append(feature)
                    if 'value' in conf['style']:
                        dense.append(feature)
                    if 'embedding' in conf['style']:
                        if f_num >=1:
                            if f_multi['combiner'] == 'none':
                                if f_multi['same']:
                                    deep.update({feature:
                                        tf.get_variable(
                                            initializer=get_initializer(**feature_initializer),
                                            shape=[1, embed_dim + first_order],
                                            name='{}'.format(feature_name))})
                                else:
                                    deep.update({feature: tf.get_variable(
                                        initializer=get_initializer(**feature_initializer),
                                        shape=[f_num, embed_dim + first_order],
                                        name='{}'.format(feature_name))})
                                con_num += f_num
                                all_num += f_num
                                con_deep_num += f_num
                                deep_num += f_num
                                deep_dim += (embed_dim+first_order) * f_num
                                dense_tag += [1]*f_num
                            else:
                                con_num += 1
                                all_num += 1
                                if f_multi['same']:
                                    deep.update({feature: tf.get_variable(
                                        initializer=get_initializer(**feature_initializer),
                                        shape=[1, embed_dim + first_order],
                                        name='{}'.format(feature_name))})
                                else:
                                    deep.update({feature: tf.get_variable(
                                        initializer=get_initializer(**feature_initializer),
                                        shape=[f_num, embed_dim + first_order],
                                        name='{}'.format(feature_name))})
                                con_deep_num += 1
                                deep_num += 1
                                deep_dim += embed_dim+first_order
                                dense_tag.append(1)

                        else:
                            if f_multi['combiner'] == 'none':
                                f_num = -f_num
                                if f_multi['same']:
                                    deep.update({feature:
                                                     tf.get_variable(
                                                         initializer=get_initializer(**feature_initializer),
                                                         shape=[1, embed_dim + first_order],
                                                         name='{}_embedding'.format(feature_name))})
                                else:
                                    deep.update({feature: tf.get_variable(initializer=get_initializer(**feature_initializer),
                                                                          shape=[f_num, embed_dim+first_order],
                                                                          name='{}_embedding'.format(feature))})
                                con_num += f_num
                                all_num += f_num
                                con_deep_num += f_num
                                deep_num += f_num
                                deep_dim += (embed_dim+first_order) * f_num
                                dense_tag += [1]*f_num
                            else:
                                f_num -= f_num
                                con_num += 1
                                all_num += 1
                                if f_multi['same']:
                                    deep.update({feature: tf.get_variable(initializer=get_initializer(**feature_initializer),
                                                                          shape=[1, embed_dim+first_order],
                                                                          name='{}_embedding'.format(feature))})
                                else:
                                    deep.update({feature: tf.get_variable(initializer=get_initializer(**feature_initializer),
                                                                          shape=[f_num, embed_dim+first_order],
                                                                          name='{}_embedding'.format(feature))})
                                con_deep_num += 1
                                deep_num += 1
                                deep_dim += embed_dim+first_order
                                dense_tag.append(1)

                    f_dim = -1
                    default_value = 0   # no used
                    if 'fill' in f_param:
                        fill_value = f_param['fill']
                        if isinstance(fill_value, str):
                            fill_value = tf.get_variable(initializer=get_initializer(**feature_initializer),
                                                         shape=[],
                                                         name='{}_fill_value'.format(fill_value))
                    else:
                        fill_value = ''
                    if 'tail' in f_param:
                        tail_value = f_param['tail']
                        if isinstance(tail_value, str):
                            tail_value = tf.get_variable(initializer=get_initializer(**feature_initializer),
                                                         shape=[],
                                                         name='{}_tail_value'.format(tail_value))
                    else:
                        tail_value = 0
                else:
                    assert False, "cant't handle this type now: {}".format(f_type)
                multi.update(
                    {feature: (f_type, f_multi['num'], f_size, f_multi['combiner'], f_multi['same'], default_value, fill_value, tail_value)}
                )

        dims = {'deep_num': deep_num, 'deep_dim': deep_dim, 'wide_dim': wide_dim, 'con_num': con_num, 'cate_num': cate_num,
                's_embed_size': embed_dim, 'cate_deep_num': deep_num-con_deep_num,
                'd_embed_size': embed_dim, 'all_num': all_num, 'dense_tag': dense_tag, 'con_deep_num': con_deep_num}
        columns = {'table': table, 'sparse': sparse, 'deep': deep, 'dense': dense, 'numeric': numeric, 'dense_tag': dense_tag, 'multi': multi}
        # dense_tag = tf.constant(dense_tag)
        return model_struct, columns, dims

    @staticmethod
    def _secondary_parse_fn(sparse_emb, deep_emb, dense_emb, mask, model_struct):
        sparse_features = tf.concat([tf.reshape(embeddings, [-1, reduce(lambda x, y:x*y, embeddings.get_shape().as_list()[1:])]) for name, embeddings in sparse_emb.items()], axis=1)
        deep_features = tf.concat([embeddings for name, embeddings in deep_emb.items()], axis=1)
        dense_features = tf.concat([embeddings for name, embeddings in dense_emb.items()], axis=1)
        return sparse_features, deep_features, dense_features

    def build_columns_information(self, params):
        self.model_struct, *Features = baseModel.build_embedding(params, self.num_shards)
        return Features

    def get_train_op_fn(self, loss, params):
        global_step = tf.compat.v1.train.get_global_step()
        learning_rate = tf.compat.v1.train.exponential_decay(self.model_conf['learning_rate'], global_step,
                                                             self.model_conf['lr_decay_step'], 0.96, staircase=True)
        if self.model_conf['optimizer'].lower() == 'adadelta':
            optimizer = tf.compat.v1.train.AdadeltaOptimizer
        elif self.model_conf['optimizer'].lower() == 'sgd':
            optimizer = tf.compat.v1.train.GradientDescentOptimizer
        elif self.model_conf['optimizer'].lower() == 'adam':
            optimizer = tf.compat.v1.train.AdamOptimizer
        elif self.model_conf['optimizer'].lower() == 'rmsprop':
            optimizer = tf.compat.v1.train.RMSPropOptimizer
        elif self.model_conf['optimizer'].lower() == 'ftrl':
            optimizer = tf.compat.v1.train.FtrlOptimizer
        else:
            assert False, '??? {} is what optimizer?'.format(self.model_conf['optimizer'])

        train_op = optimizer(learning_rate=learning_rate, beta1=0.9, beta2=0.999,
                                                    epsilon=1e-8).minimize(loss, global_step=global_step)
        return train_op

    def get_loss(self, labels, predictions):
        labels = tf.cast(labels, tf.int32)
        # labels = tf.expand_dims(labels, axis=1)
        l1_reg, l2_reg = self.model_conf['l1_reg'], self.model_conf['l2_reg']
        loss = tf.compat.v1.losses.log_loss(labels, predictions)
        T_vars = tf.compat.v1.trainable_variables()
        M_vars = [var for var in T_vars if var.name.startswith('trunk')]
        E_vars = list(set(T_vars)-set(M_vars))
        l1_loss, l2_loss = 0, 0
        if l2_reg > 0:
            for vars in T_vars:
                l2_loss += tf.keras.regularizers.l2(
                    0.5 * (l2_reg))(vars)
        if l1_reg > 0:
            for vars in M_vars:
                l1_loss += tf.keras.regularizers.l1(
                    l1_reg)(vars)
        loss = loss + l1_loss + l2_loss
        return loss

    @abstractmethod
    def get_eval_metric_ops(self, labels, predictions):
        pass

    def forward(self, indexed_input, params, is_training):
        columns, dims = self.build_columns_information(params)
        self.dims, model_input = self.embedding_parser_fn(indexed_input, columns, dims, self.model_struct)
        out = self._forward_model(is_training, *model_input)
        if self.mode != 'train':
            for out_node in self.out_node_names:
                try:
                    _ = tf.get_default_graph().get_operation_by_name(out_node)
                except:
                    raise LookupError('Cant find out node:{} in Graph'.format(out_node))
        return out

    @abstractmethod
    def _forward_model(self, is_training, *args, **kwargs):
        '''
        :param sparse_features: [batch_size, len of all one hot merge], one_hot, only from category field generally
        :param deep_features: [batch_size, [cate, con]num, embedding_dim], embedding, concat of category field and numeric
        :param dense_features: [batch_size, len of con], concat real value of numeric data
        :param is_training: bool
        :return:
        '''
        pass

    def add_summary(self, labels, predictions, eval_metric_ops):
        labels = tf.cast(labels, tf.float32)
        precision = tf.reduce_mean(
            input_tensor=tf.cast(
                tf.less(tf.abs(predictions - labels), 0.5),
                tf.float32))
        tf.compat.v1.summary.scalar('precision', precision)
        tf.compat.v1.summary.scalar('auc_train', eval_metric_ops['auc'][1])

    def get_predictions_out(self, features, predictions, pids, out_format):
        predictions_out = {id:tf.expand_dims(tf.identity(pids[id], name=id), axis=1) if id[:-1] not in self.out_node_names else predictions
                           for id in out_format}
        return predictions_out


def get_initializer(name=None, param=None):
    param = dict() if param is None else param
    try:
        return eval(name)(**param)
    except (TypeError, NameError, AttributeError):
        return tf.keras.initializers.he_normal()



import tensorflow as tf
from .BaseModel import baseModel, get_initializer
from functools import reduce
from operator import mul
try:
    from dl_rank.layer import layers
except:
    from layer import layers

VERY_BIG_NUMBER = 1e30
VERY_SMALL_NUMBER = 1e-30
VERY_POSITIVE_NUMBER = VERY_BIG_NUMBER
VERY_NEGATIVE_NUMBER = -VERY_BIG_NUMBER


# wide columns

class _config():
    def __init__(self, cfg):
        self.reduce_sequence_length = cfg['reduce_sequence_length']
        self.embedding_dim = cfg['reduce_sequence_length']
        self.dropout = cfg['dropout']
        self.last_dropout = cfg['last_dropout']
        self.hidden_num = cfg['hidden_num']
        self.wd = cfg['wd']
        self.out_node_name = ['predict_value']
        self.init_std = cfg['init_std']



class wide_and_deep(baseModel):
    name = 'wide_deep_traditional_attention'
    def __init__(self, model_conf, mode):
        super(wide_and_deep, self).__init__(model_conf, mode)
        self.optimizer_name = model_conf['optimizer']
        self.l1 = model_conf['l1_reg']
        self.l2 = model_conf['l2_reg']
        self.deep_lr = model_conf['deep_learning_rate']
        self.lr = model_conf['learning_rate']

        # self.use_first_order = self.model_conf['first_order']
        self.dnn_hidden_units = self.model_conf['dnn_hidden_units']

        self.batch_size = model_conf['batch_size']
        self.cfg = _config(model_conf['cfg'])
        self.mode = mode

        self.init_std = self.cfg.init_std

    def _forward_model(self, is_training, *args, **kwargs):
        # sparse_features, deep_features, dense_features = _args
        self.is_train = is_training
        with tf.variable_scope('wide_layer_variable') as scope:
            self.b = tf.get_variable(initializer=get_initializer(param={'stddev': 0.01}), shape=[], name='bias')
            # self.b = tf.Variable(self.init_std, trainable=True, name='bias')
        with tf.variable_scope('trunk'):
            """Wide_layer is a basic LR model, including click order, transaction order,
            item statistical characteristics and query pair pairs. The deep model connects
            to two traditional attention models, and finally adds the output layer linearly.
            """
            if self.mode == 'train':
                sparse_features, itemid_emb, sequence_emb, sequence_token_mask, itemid_token_mask = args
                # tfprint = tf.print(*[tf.shape(t) for t in args], 'shape', summarize=10000)
                # with tf.control_dependencies([]):
                #     sparse_features = tf.identity(sparse_features)
                # wide_layer
                # deep_layer
                # sequence_emb = tf.random.normal([self.batch_size, 20, self.cfg.embedding_dim])
                # itemid_emb = tf.random.normal([self.batch_size, self.embedding_size])
                # sequence_token_mask = tf.random.normal([self.batch_size, self.cfg.reduce_sequence_length])
                # itemid_token_mask = tf.random.normal([self.batch_size, 1])
            else:
                sparse_features, itemid_emb, sequence_emb, _, _ = args
                self.sequence = tf.identity(sequence_emb, name="seq_input_node")                    # placeholder: embedding_lookup_2/Reshape_1
                self.itemid = tf.squeeze(itemid_emb, axis=1, name="click_input_node")    # placeholder: trunk/Squeeze

                # self.sequence = tf.random.normal([None, self.cfg.reduce_sequence_length, self.cfg.embedding_dim], name='sequence')
                # self.itemid = tf.random.normal([None, self.cfg.embedding_dim], name='itemid')
                sequence_expand = tf.reduce_mean(self.sequence, 2)
                sequence_token_mask = tf.cast(sequence_expand, tf.bool)
                itemid_token_mask = tf.cast(tf.reduce_mean(tf.expand_dims(self.itemid, 1), 2), tf.bool)
            with tf.variable_scope("wide_layer") as scope:
                if self.mode == 'train':
                    self.add_sum = tf.reduce_sum(sparse_features, axis=1, name='wide_value')
                    self.logits_wide =  self.b # self.add_sum +
                    self.ctr = tf.sigmoid(self.logits_wide)
                    print('logits_wide_shape', self.logits_wide.get_shape())

            with tf.variable_scope("deep_layer") as scope:
                if self.mode == 'train':
                    # sequence_emb = tf.nn.embedding_lookup(self.item_mat, tf.slice(self.sequence, [0, 0], [self.batch_size,
                    #                                                                                       self.cfg.reduce_sequence_length]))
                    # itemid_emb = tf.nn.embedding_lookup(self.item_mat, self.itemid)
                    # sequence_token_mask = tf.cast(
                    #     tf.slice(self.sequence, [0, 0], [self.batch_size, self.cfg.reduce_sequence_length]), tf.bool)
                    # itemid_token_mask = tf.cast(self.itemid, tf.bool)
                    self.logits_deep = self.bulid_attention_layers(is_training, sequence_emb, sequence_token_mask, itemid_emb,
                                                                   itemid_token_mask, 'attention')
                else:
                    # deep input
                    self.bs = tf.shape(self.itemid)[0]
                    print('bs_shape', self.bs)
                    self.logits_deep = self.bulid_attention_layers(is_training, self.sequence, sequence_token_mask,
                                                                   tf.expand_dims(self.itemid, 1),
                                                                   itemid_token_mask, 'attention')
                # self.logits_deep=tf.squeeze(logits_deep)
                print('logits_deep_shape', self.logits_deep.get_shape())
            # self.tfprint = tf.print('logits_deep', self.logits_deep, summarize=10000)

            with tf.variable_scope("output_layer"):
                if self.mode == 'train':

                    # with tf.control_dependencies([self.tfprint]):
                    self.logits = self.logits_wide + self.logits_deep
                    tf.summary.histogram('logits_wide', self.logits_wide)
                    tf.summary.histogram('logits_deep', self.logits_deep)
                    tf.summary.histogram('logits', self.logits)
                else:
                    self.logits = self.logits_deep
                print('logits_shape', self.logits.get_shape())
            # self.y = tf.identity(self.logits_deep, name=self.out_node_names[0])
            self.predictions = tf.nn.sigmoid(self.logits)
        self.y = tf.identity(self.logits_deep, name=self.out_node_names[0])
        return self.predictions

    @staticmethod
    def _secondary_parse_fn(sparse_emb, deep_emb, dense_emb, mask, model_struct):
        wide = model_struct['wide']
        itemid = model_struct['trigger_id'][0]
        sequence = model_struct['seq'][0]
        # tfprint = tf.print('cid', deep_emb['cid'], 'click_cross', deep_emb['click_cross'], 'order', deep_emb['order_cross'])
        with tf.control_dependencies([]):
            sparse_features = tf.concat([tf.squeeze(deep_emb[w], axis=1) for w in wide], axis=1)
        itemid_emb = deep_emb[itemid]
        sequence_emb = deep_emb[sequence]
        sequence_token_mask = mask[sequence]
        itemid_token_mask = tf.ones([tf.shape(mask[sequence])[0], 1])
        return sparse_features, itemid_emb, sequence_emb, sequence_token_mask, itemid_token_mask

    def get_eval_metric_ops(self, labels, predictions):
        """Return a dict of the evaluation Ops.
        Args:
            labels (Tensor): Labels tensor for training and evaluation.
            predictions (Tensor): Predictions Tensor.
        Returns:
            Dict of metric results keyed by name.
        """
        # labels = tf.expand_dims(tf.cast(labels, tf.int32), axis=1)
        labels = tf.cast(labels, tf.int32)
        auc = tf.compat.v1.metrics.auc(
            labels=labels,
            predictions=predictions,
            name='auc',
            curve='ROC',
        )

        return {
            'auc': auc
        }

    def bulid_optimize(self,learning_rate):

        # ---------- optimization ---------
        if self.optimizer_name.lower() == 'adadelta':
            opt = tf.train.AdadeltaOptimizer(learning_rate)
        elif self.optimizer_name.lower() == 'sgd':
            opt = tf.train.GradientDescentOptimizer(learning_rate)
        elif self.optimizer_name.lower() == 'adam':
            opt = tf.train.AdamOptimizer(learning_rate)
        elif self.optimizer_name.lower() == 'rmsprop':
            opt = tf.train.RMSPropOptimizer(learning_rate)
        elif self.optimizer_name.lower() == 'ftrl':
            opt= tf.train.FtrlOptimizer(learning_rate=learning_rate,
                                               initial_accumulator_value=1.0 / pow(self.batch_size, 2),
                                               l1_regularization_strength=self.l1 / self.batch_size,
                                               l2_regularization_strength=self.l2 / self.batch_size,
                                               use_locking=True)
        elif self.optimizer_name.lower() == 'combine':
            wide_opt=tf.train.FtrlOptimizer(learning_rate=learning_rate,
                                               initial_accumulator_value=1.0 / pow(self.batch_size, 2),
                                               l1_regularization_strength=self.l1 / self.batch_size,
                                               l2_regularization_strength=self.l2 / self.batch_size,
                                               use_locking=True)

            #deep_opt = tf.contrib.opt.LazyAdamOptimizer(learning_rate=0.0004, beta1=0.9, beta2=0.999)
            deep_opt = tf.train.FtrlOptimizer(learning_rate=learning_rate,initial_accumulator_value=1.0 / pow(self.batch_size, 2),
                                              l1_regularization_strength=0.3/self.batch_size,
                                              l2_regularization_strength=5.0/self.batch_size,
                                              use_locking=True)

            opt=[wide_opt,deep_opt]
        elif self.optimizer_name.lower() == 'ftrl_adadelta':
            wide_opt=tf.train.FtrlOptimizer(learning_rate=learning_rate,
                                               initial_accumulator_value=1.0 / pow(self.batch_size, 2),
                                               l1_regularization_strength=self.l1 / self.batch_size,
                                               l2_regularization_strength=self.l2 / self.batch_size,
                                               use_locking=True)

            deep_opt = tf.train.AdadeltaOptimizer(learning_rate= 1.0)

            opt=[wide_opt,deep_opt]

        elif self.optimizer_name.lower() == 'ftrl_adam':
            wide_opt = tf.train.FtrlOptimizer(learning_rate=learning_rate,
                                              initial_accumulator_value=1.0 / pow(self.batch_size, 2),
                                              l1_regularization_strength=self.l1 / self.batch_size,
                                              l2_regularization_strength=self.l2 / self.batch_size,
                                              use_locking=True)

            deep_opt = tf.contrib.opt.LazyAdamOptimizer(learning_rate=self.deep_lr, beta1=0.9, beta2=0.999)

            opt = [wide_opt, deep_opt]
        else:
            raise AttributeError('no optimizer named as \'%s\'' % self.optimizer_name)
        return opt

    def get_train_op_fn(self, loss, params):
        self.global_step = tf.compat.v1.train.get_global_step()
        self.loss = loss

        self.optimizer = self.bulid_optimize(self.lr)

        if self.optimizer_name.lower() in ['combine', 'ftrl_adadelta', 'ftrl_adam']:
            self.trainer = []
            self.wide_optimizer, self.deep_optimizer = self.optimizer[0], self.optimizer[1]
            wide_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, "wide_layer")
            print('trainable wide var num: %d' % len(wide_vars))
            deep_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, "deep_layer")
            print('trainable deep var num: %d' % len(deep_vars))
            self.trainer.append(self.wide_optimizer.minimize(self.loss, self.global_step, var_list=wide_vars))
            "Cross optimization"
            self.trainer.append(
                self.deep_optimizer.minimize(self.loss, self.global_step, var_list=deep_vars))
            """
            self.trainer.append(
                self.deep_optimizer.minimize(self.loss, self.global_step, var_list=deep_vars))
            """
            self.trainer = tf.group(*self.trainer)
        else:
            self.trainer = self.optimizer.minimize(self.loss, global_step=self.global_step)
        return self.trainer

    def bulid_attention_layers(self, is_train, u_emb, all_token_mask, itemid_emb, temid_token_mask, scope):
        # partitioner = tf.fixed_size_partitioner(self.partition)

        with tf.variable_scope('attention'):#, partitioner=partitioner):
            if not self.mode == 'train':
                sequence_rep_single = single_attention(
                    u_emb, all_token_mask, scope + '_sequence', self.cfg.dropout, is_train, self.cfg.wd,
                    'elu', None, 's1_attention', is_multi_att=False, attention_dim=None
                )
                sequence_rep = tf.tile(sequence_rep_single, multiples=[self.bs, 1])
                tf.get_variable_scope().reuse_variables()
                itemid_rep = single_attention(
                    itemid_emb, temid_token_mask, scope + '_sequence', self.cfg.dropout, is_train, self.cfg.wd,
                    'elu', None, 's2_attention', is_multi_att=False, attention_dim=None
                )
            else:
                sequence_rep = single_attention(
                    u_emb, all_token_mask, scope + '_sequence', self.cfg.dropout, is_train, self.cfg.wd,
                    'elu', None, 's1_attention', is_multi_att=False, attention_dim=None
                )
                tf.get_variable_scope().reuse_variables()
                itemid_rep = single_attention(
                    itemid_emb, temid_token_mask, scope + '_sequence', self.cfg.dropout, is_train, self.cfg.wd,
                    'elu', None, 's2_attention', is_multi_att=False, attention_dim=None
                )

            out_rep = tf.concat([sequence_rep, itemid_rep, sequence_rep - itemid_rep, sequence_rep * itemid_rep],
                                1)  # [bs,128*4]
        with tf.variable_scope('output_layer_last'):#, partitioner=partitioner):
            attention_layer = dice(
                linear([out_rep], self.cfg.hidden_num, True, 0., scope=scope + '_output_attention_layer', squeeze=False,
                       wd=self.cfg.wd, input_keep_prob=self.cfg.dropout, is_train=self.is_train),
                name='attention_pre_out')

            # logits_deep_dropout = tf.nn.dropout(attention_layer,self.cfg.last_dropout)
            # logits_deep = tf.layers.dense(inputs=logits_deep_dropout, units=1, activation=None, use_bias=True)[:,0]
            logits_deep = linear([attention_layer], 1, True, 0., scope=scope + '_logits', squeeze=False,
                                 wd=self.cfg.wd, input_keep_prob=self.cfg.last_dropout, is_train=self.is_train)[:, 0]
        return logits_deep

    def get_loss(self, labels, predictions):
        """Using sigmoid_cross_entropy_with_logits, the shape constraints of label and Logits are consistent and one-dimensional array."""

        # trainable_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES,'deep_layer')
        # print('trainable var num: %d' % len(trainable_vars))
        labels = tf.cast(labels, tf.float32, name='true_label')
        labels = tf.squeeze(labels, axis=1)
        labels = tf.slice(labels, [0], tf.shape(self.logits))
        losses = tf.nn.sigmoid_cross_entropy_with_logits(labels=labels, logits=self.logits)
        loss = tf.reduce_mean(losses, name='loss')
        tf.summary.scalar('loss', loss)

        # weights = tf.slice(self.reweight, [0], tf.shape(self.logits))
        self.loss_sum = tf.reduce_sum(losses, name='loss_sum')
        # self.weight_sum = tf.reduce_sum(weights, name='weight_sum')
        tf.summary.scalar('loss_sum', self.loss_sum)
        # tf.summary.scalar('weight_sum', self.weight_sum)

        # cid_tensor = tf.get_default_graph().get_tensor_by_name("click:0")[0]
        # w_tensor = tf.get_default_graph().get_tensor_by_name("trunk/deep_layer/output_layer_last/attention_output_attention_layer/bias:0")
        # gradients_node = tf.gradients(loss, w_tensor)
        # click_emb_tensor = tf.get_default_graph().get_tensor_by_name("embedding_lookup_2/Identity:0")[0]
        # tp = tf.print(cid_tensor, 'id', click_emb_tensor, "click", summarize=10000)
        # with tf.control_dependencies([tp]):
        #     loss = tf.identity(loss)
        return loss

    def add_summary(self, labels, predictions, eval_metric_ops):
        labels = tf.cast(labels, tf.float32)
        b = tf.reshape(tf.concat([labels, predictions], axis=1), [-1])
        # tp = tf.print(b, summarize=1e3)
        # with tf.control_dependencies([]):
        precision = tf.reduce_mean(
            input_tensor=tf.cast(
                tf.less(tf.abs(predictions - labels), 0.5),
                tf.float32))
        tf.compat.v1.summary.scalar('precision', precision)
        tf.compat.v1.summary.scalar('auc_train', eval_metric_ops['auc'][1])

def single_attention(rep_tensor, rep_mask, scope=None,
                    keep_prob=1., is_train=None, wd=0., activation='elu',
                    tensor_dict=None, name='',is_multi_att=False, attention_dim=None):

    """single attention: contain two traditional attention struct.

    Args:
      rep_tensor: list tensor,shape is [batch_size,list_size,embedding_dim].
      rep_mask: whether the marker Tensor is 0,bool value, shape is [batch_size,list_size]
    Returns:
      Attention representation of tensor,shape is [batch_size,embedding_dim]
    Raises:
      TypeError: If the input dimension is incorrect.
    """

    with tf.variable_scope(scope):
        attention_rep_first_layer = traditional_attention(
            rep_tensor, rep_mask, 'traditional_attention',
            keep_prob, is_train, wd, activation,
            tensor_dict=tensor_dict, name=name + '_attn')

        #attention_rep_final = traditional_attention(
        #    attention_rep_first_layer, rep_mask, 'traditional_attention',
        #    keep_prob, is_train, wd, activation,
        #    tensor_dict=tensor_dict, name=name + '_attn')

        return attention_rep_first_layer

def get_logits(args, size, bias, bias_start=0.0, scope=None, mask=None, wd=0.0,
               input_keep_prob=1.0, is_train=None, func=None):
    if func is None:
        func = "linear"
    if func == 'sum':
        return sum_logits(args, mask=mask, name=scope)
    elif func == 'linear':
        return linear_logits(args, bias, bias_start=bias_start, scope=scope, mask=mask, wd=wd, input_keep_prob=input_keep_prob,
                             is_train=is_train)
    elif func == 'dot':
        assert len(args) == 2
        arg = args[0] * args[1]
        return sum_logits([arg], mask=mask, name=scope)
    elif func == 'mul_linear':
        assert len(args) == 2
        arg = args[0] * args[1]
        return linear_logits([arg], bias, bias_start=bias_start, scope=scope, mask=mask, wd=wd, input_keep_prob=input_keep_prob,
                             is_train=is_train)
    elif func == 'proj':
        assert len(args) == 2
        d = args[1].get_shape()[-1]
        proj = linear([args[0]], d, False, bias_start=bias_start, scope=scope, wd=wd, input_keep_prob=input_keep_prob,
                      is_train=is_train)
        return sum_logits([proj * args[1]], mask=mask)
    elif func == 'tri_linear':
        assert len(args) == 2
        new_arg = args[0] * args[1]
        return linear_logits([args[0], args[1], new_arg], bias, bias_start=bias_start, scope=scope, mask=mask, wd=wd, input_keep_prob=input_keep_prob,
                             is_train=is_train)
    else:
        raise Exception()

def dice(_x, axis=-1, epsilon=0.0000001, name=''):

    alphas = tf.get_variable('alpha' + name, _x.get_shape()[-1],
                             initializer=tf.constant_initializer(0.0),
                             dtype=tf.float32)

    input_shape = list(_x.get_shape())
    reduction_axes = list(range(len(input_shape)))

    del reduction_axes[axis]  # [0]

    broadcast_shape = [1] * len(input_shape)  # [1,1]
    broadcast_shape[axis] = input_shape[axis]  # [1 * hidden_unit_size]

    # case: train mode (uses stats of the current batch)
    mean = tf.reduce_mean(_x, axis=reduction_axes)  # [1 * hidden_unit_size]
    brodcast_mean = tf.reshape(mean, broadcast_shape)
    std = tf.reduce_mean(tf.square(_x - brodcast_mean) + epsilon, axis=reduction_axes)
    std = tf.sqrt(std)
    brodcast_std = tf.reshape(std, broadcast_shape)  # [1 * hidden_unit_size]
    # x_normed = (_x - brodcast_mean) / (brodcast_std + epsilon)
    x_normed = tf.layers.batch_normalization(_x, center=False, scale=False)  # a simple way to use BN to calculate x_p
    x_p = tf.sigmoid(x_normed)

    return alphas * (1.0 - x_p) * _x + x_p * _x

def traditional_attention(rep_tensor, rep_mask, scope=None,
                          keep_prob=1., is_train=None, wd=0., activation='elu',
                          tensor_dict=None, name=None,output_dim=None):

    bs, sl, vec = tf.shape(rep_tensor)[0], tf.shape(rep_tensor)[1], tf.shape(rep_tensor)[2]
    if output_dim is None:
        ivec = rep_tensor.get_shape()[2]
    else:
        ivec = output_dim
    with tf.variable_scope(scope):
        rep_tensor_map = bn_dense_layer(rep_tensor, ivec, True, 0., 'bn_dense_map', activation,
                                        False, wd, keep_prob, is_train)
        rep_tensor_logits = get_logits([rep_tensor_map], None, False, scope='self_attn_logits',
                                       mask=rep_mask, input_keep_prob=keep_prob, is_train=is_train)  # bs,sl
        attn_res = softsel(rep_tensor, rep_tensor_logits, rep_mask)  # bs,vec

        # save attn
        #if tensor_dict is not None and name is not None:
        #    tensor_dict[name] = tf.nn.softmax(rep_tensor_logits)

        return attn_res

def bn_dense_layer(input_tensor, hn, bias, bias_start=0.0, scope=None,
                   activation='relu', enable_bn=True,
                   wd=0., keep_prob=1.0, is_train=None):
    if is_train is None:
        is_train = False

    # activation
    if activation == 'linear':
        activation_func = tf.identity
    elif activation == 'relu':
        activation_func = tf.nn.relu
    elif activation == 'elu':
        activation_func = tf.nn.elu
    elif activation == 'selu':
        activation_func = selu
    else:
        raise AttributeError('no activation function named as %s' % activation)

    with tf.variable_scope(scope or 'bn_dense_layer'):
        linear_map = linear(input_tensor, hn, bias, bias_start, 'linear_map',
                            False, wd, keep_prob, is_train)
        if enable_bn:
            with tf.variable_scope("bn", reuse=True):
                linear_map = tf.layers.batch_normalization(linear_map, center=True, scale=True, training=is_train)
        return activation_func(linear_map)


def softsel(target, logits, mask=None, scope=None):
    """
    :param target: [ ..., J, d] dtype=float #(b,sn,sl,ql,d)
    :param logits: [ ..., J], dtype=float
    :param mask: [ ..., J], dtype=bool
    :param scope:
    :return: [..., d], dtype=float
    """
    with tf.name_scope(scope or "Softsel"):
        a = softmax(logits, mask=mask)
        target_rank = len(target.get_shape().as_list())
        out = tf.reduce_sum(tf.expand_dims(a, -1) * target, target_rank - 2)
        return out

def sum_logits(args, mask=None, name=None):
    with tf.name_scope(name or "sum_logits"):
        if args is None or (isinstance(args, (tuple, list)) and not args):
            raise ValueError("`_args` must be specified")
        if not isinstance(args, (tuple, list)):
            args = [args]
        rank = len(args[0].get_shape())
        logits = sum(tf.reduce_sum(arg, rank - 1) for arg in args)
        if mask is not None:
            logits = exp_mask(logits, mask)
        return logits

def softmax(logits, mask=None, scope=None):
    with tf.name_scope(scope or "Softmax"):
        if mask is not None:
            logits = exp_mask(logits, mask)
        out = tf.nn.softmax(logits,-1)
        return out

def linear(args, output_size, bias, bias_start=0.0, scope=None, squeeze=False, wd=0.0, input_keep_prob=1.0,
           is_train=None):
    if args is None or (isinstance(args, (tuple, list)) and not args):
        raise ValueError("`_args` must be specified")
    if not isinstance(args, (tuple, list)):
        args = [args]

    flat_args = [flatten(arg, 1) for arg in args] # for dense layer [(-1, d)]
    if input_keep_prob < 1.0:
        assert is_train is not None
        flat_args = [tf.layers.dropout(arg, rate=input_keep_prob, training=is_train) for arg in flat_args]
        # flat_args = [tf.cond(is_train, lambda: tf.nn.dropout(arg, input_keep_prob), lambda: arg)# for dense layer [(-1, d)]
        #              for arg in flat_args]
    flat_out = _linear(flat_args, output_size, bias, bias_start=bias_start, scope=scope) # dense
    out = reconstruct(flat_out, args[0], 1) # ()
    if squeeze:
        out = tf.squeeze(out, [len(args[0].get_shape().as_list())-1])

    if wd:
        add_reg_without_bias()

    return out

def flatten(tensor, keep):
    fixed_shape = tensor.get_shape().as_list()
    start = len(fixed_shape) - keep
    left = reduce(mul, [fixed_shape[i] or tf.shape(tensor)[i] for i in range(start)])
    out_shape = [left] + [fixed_shape[i] or tf.shape(tensor)[i] for i in range(start, len(fixed_shape))]
    flat = tf.reshape(tensor, out_shape)
    return flat

def _linear(xs,output_size,bias,bias_start=0., scope=None):
    with tf.variable_scope(scope or 'linear_layer'):
        x = tf.concat(xs,-1)
        input_size = x.get_shape()[-1]
        W = tf.get_variable('W', shape=[input_size,output_size],dtype=tf.float32,)
        if bias:
            bias = tf.get_variable('bias', shape=[output_size],dtype=tf.float32,
                                   initializer=tf.constant_initializer(0))
            out = tf.matmul(x, W) + bias
        else:
            out = tf.matmul(x, W)

        return out


def add_reg_without_bias(scope=None):
    pass
    # scope = scope or tf.get_variable_scope().name
    # variables = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope=scope)
    # counter = 0
    # for var in variables:
    #     if len(var.get_shape().as_list()) <= 1: continue
    #     tf.add_to_collection('reg_vars', var)
    #     counter += 1
    # return counter

def reconstruct(tensor, ref, keep, dim_reduced_keep=None):
    dim_reduced_keep = dim_reduced_keep or keep

    ref_shape = ref.get_shape().as_list() # original shape
    tensor_shape = tensor.get_shape().as_list() # current shape
    ref_stop = len(ref_shape) - keep # flatten dims list
    tensor_start = len(tensor_shape) - dim_reduced_keep  # start
    pre_shape = [ref_shape[i] or tf.shape(ref)[i] for i in range(ref_stop)] #
    keep_shape = [tensor_shape[i] or tf.shape(tensor)[i] for i in range(tensor_start, len(tensor_shape))] #
    # pre_shape = [tf.shape(ref)[i] for i in range(len(ref.get_shape().as_list()[:-keep]))]
    # keep_shape = tensor.get_shape().as_list()[-keep:]
    target_shape = pre_shape + keep_shape
    out = tf.reshape(tensor, target_shape)
    return out

def selu(x):
    with tf.name_scope('elu') as scope:
        alpha = 1.6732632423543772848170429916717
        scale = 1.0507009873554804934193349852946
        return scale*tf.where(x>=0.0, x, alpha*tf.nn.elu(x))

def linear_logits(args, bias, bias_start=0.0, scope=None, mask=None, wd=0.0, input_keep_prob=1.0, is_train=None):
    with tf.variable_scope(scope or "Linear_Logits"):
        logits = linear(args, 1, bias, bias_start=bias_start, squeeze=True, scope='first',
                        wd=wd, input_keep_prob=input_keep_prob, is_train=is_train)
        if mask is not None:
            logits = exp_mask(logits, mask)
        return logits

def exp_mask(val, mask, name=None):
    """Give very negative number to unmasked elements in val.
    For example, [-3, -2, 10], [True, True, False] -> [-3, -2, -1e9].
    Typically, this effectively masks in exponential space (e.g. softmax)
    Args:
        val: values to be masked
        mask: masking boolean tensor, same shape as tensor
        name: name for output tensor

    Returns:
        Same shape as val, where some elements are very small (exponentially zero)
    """
    if name is None:
        name = "exp_mask"
    return tf.add(val, (1 - tf.cast(mask, 'float')) * VERY_NEGATIVE_NUMBER, name=name)




import os
# from pyspark import SparkFiles
import boto3
from functools import reduce
import subprocess
import pyhdfs
import sys

class hdfsFile(object):
    ip = ''
    port = ''
    client = None

    @staticmethod
    def init_client(ip_port):
        if 'HADOOP_HOME' in os.environ and 'HADOOP_HDFS_HOME' in os.environ:
            hdfsFile._copy_libhdfs()
            classpath = subprocess.check_output('hadoop classpath --glob', shell=True).decode('utf-8').strip()
            if 'CLASSPATH' not in os.environ:
                os.environ['CLASSPATH'] = ''
            if classpath not in os.environ['CLASSPATH']:
                os.environ['CLASSPATH'] = ':'.join([os.environ['CLASSPATH'], classpath])
            if ':' in ip_port:
                hdfsFile.ip, hdfsFile.port = ip_port.split(':')
                hdfsFile.client = pyhdfs.HdfsClient(ip_port)
        else:
            pass

    @staticmethod
    def _copy_libhdfs():
        src = os.path.join(os.environ['HADOOP_HOME'], 'lib/native/libhdfs.so')
        dst_dir = os.path.join(os.environ['HADOOP_HDFS_HOME'], 'lib/native')
        if not os.path.exists(dst_dir):
            os.system('sudo mkdir {dir}'.format(dir=dst_dir))
            os.system('sudo ln -s {src_path} {dst_path}'.format(src_path=src, dst_path=os.path.join(dst_dir, 'libhdfs.so')))

    @staticmethod
    def DeleteRecursively(path):
        hdfsFile.client.delete(path, recursive=True)

    @staticmethod
    def Remove(path):
        hdfsFile.client.delete(path)

    @staticmethod
    def ListDirectory(dirname):
        return hdfsFile.client.listdir(dirname)

    @staticmethod
    def CopyLocal(srcPath, dstPath):
        if dstPath.startswith('hdfs://'):
            hdfsFile.client.copy_from_local(srcPath, dstPath)
        elif srcPath.startswith('hdfs://'):
            hdfsFile.client.copy_to_local(srcPath, dstPath)

    @staticmethod
    def Copy(srcPath, dstPath, overwrite=False):
        if srcPath != dstPath:
            if overwrite and hdfsFile.Exists(dstPath):
                hdfsFile.Remove(dstPath)
            else:
                os.system('hadoop fs -cp {src} {dst}'.format(src=srcPath, dst=dstPath))
    @staticmethod
    def CopyS3(srcPath, dstPath):
        cmd = 'hadoop distcp {src} {dst}'.format(src=srcPath, dst=dstPath)
        os.system(cmd)

    @staticmethod
    def MakeDirs(filePath):
        if not hdfsFile.client.exists(filePath):
            hdfsFile.client.mkdirs(filePath)

    @staticmethod
    def Rename(srcPath, dstPath):
        if not hdfsFile.client.exists(srcPath):
            return False
        os.system('hadoop fs -mv {src} {dst}'.format(src=srcPath, dst=dstPath))
        # hdfsFile.client.rename(srcPath, dstPath)
        return True

    @staticmethod
    def RenameDirectory(srcPath, dstPath):
        hdfsFile.Rename(srcPath, dstPath)

    @staticmethod
    def Walk(top):
        return hdfsFile.client.walk(top)

    @staticmethod
    def Open(filename):
        return hdfsFile.client.open(filename)

    @staticmethod
    def Exists(filename):
        return hdfsFile.client.exists(filename)

    @staticmethod
    def IsDirectory(filename):
        try:
            filestatus = hdfsFile.client.get_file_status(filename)
            if filestatus.childrenNum:
                return True
            else:
                return False
        except:
            # not exists
            return False

class s3File(object):
    s3 = boto3.resource('s3')
    client = boto3.client('s3')
    @staticmethod
    def split_s3_path(filename):
        filename = filename[len('s3://'):]
        bucket, key = filename.split('/', 1)
        return bucket, key

    @staticmethod
    def _list_obj_under_path(filename):
        bucket, keyname = s3File.split_s3_path(filename)
        bucket = s3File.s3.Bucket(bucket)
        objs = bucket.objects.filter(Prefix=keyname)
        return objs

    @staticmethod
    def Exists(filename):
        filename = removetail(filename)
        _, keyname = s3File.split_s3_path(filename)
        objs = iter(s3File._list_obj_under_path(filename))
        try:
            obj = next(objs)
            obj_tail = obj.key[len(keyname):]
            if obj_tail == '' or obj_tail.startswith('/'):
                return True
            else:
                return False
        except StopIteration:
            return False

    @staticmethod
    def IsDirectory(filename):
        filename = addtail(filename)
        objs = iter(s3File._list_obj_under_path(filename))
        try:
            _ = next(objs)
            return True
        except StopIteration:
            return False

    @staticmethod
    def ListDirectory(dirname):
        _, dirs, files = s3File.Walk(dirname, True)
        dirs = [d.strip('/') for d in dirs]
        return dirs + files

    @staticmethod
    def Open(filename):
        bucket, keyname = s3File.split_s3_path(filename)
        obj = s3File.s3.Object(bucket, keyname)
        return s3FileObj(obj.get()['Body'])#.read().decode('utf-8')

    @staticmethod
    def MakeDirs(dirname):
        bucket, keyname = s3File.split_s3_path(dirname)
        if keyname[-1] != '/':
            keyname = keyname + '/'
            s3File.client.put_object(Bucket=bucket, Key=keyname)

    @staticmethod
    def DeleteRecursively(dirname):
        obj = s3File._list_obj_under_path(dirname)
        obj.delete()

    @staticmethod
    def Walk(dirname, one_step=False):
        bucket, keyroot = s3File.split_s3_path(dirname)
        all_obj = [obj.key for obj in list(s3File._list_obj_under_path(dirname))]

        def _walk(root):
            obj_under_dir = [obj[len(root):] for obj in filter(lambda name: name.startswith(root), all_obj)]
            pre_key_under_dir = set([obj.split('/')[0]+'/' if '/' in obj else obj for obj in obj_under_dir])
            if '' in pre_key_under_dir:
                pre_key_under_dir.remove('')
            dirs, files = reduce(lambda x, y: (x[0]+[y], x[1]) if '/' in y else (x[0], x[1]+[y]),
                                 pre_key_under_dir,
                                 ([], []))
            return root, dirs, files

        def _recursive_walk(root):
            root, dirs, files = _walk(root)
            yield 's3://'+os.path.join(bucket, root), \
                  [d.strip('/') for d in dirs], \
                  files
            for d in dirs:
                yield from _recursive_walk(os.path.join(root, d))

        if keyroot[-1] != '/':
            keyroot += '/'
        if one_step:
            return _walk(keyroot)
        else:
            return _recursive_walk(keyroot)

    @staticmethod
    def Rename(oldname, newname):
        s3File.Copy(oldname, newname)
        oldbucket, oldkey = s3File.split_s3_path(oldname)
        s3File.s3.Object(oldbucket, oldkey).delete()

    @staticmethod
    def RenameDirectory(oldname, newname):
        oldname = addtail(oldname)
        newname = addtail(newname)
        if oldname != newname:
            bucket, _ = s3File.split_s3_path(oldname)
            all_obj = [obj.key for obj in list(s3File._list_obj_under_path(oldname))]
            for old_key in all_obj:
                old_key = os.path.join('s3://'+bucket, old_key)
                new_key = newname + old_key[len(oldname):]
                if old_key[-1] == '/':
                    if not s3File.Exists(new_key):
                        s3File.MakeDirs(new_key)
                else:
                    s3File.Copy(old_key, new_key)
            s3File.DeleteRecursively(oldname)


    @staticmethod
    def Remove(filename):
        bucket, key = s3File.split_s3_path(filename)
        s3File.s3.Object(bucket, key).delete()

    @staticmethod
    def CopyLocal(srcPath, dstPath, is_recursive=False):
        cmd = 'aws s3 cp {srcPath} {dstPath}'.format(srcPath=srcPath, dstPath=dstPath)
        if is_recursive:
            cmd += ' --recursive'
        os.system(cmd)

    @staticmethod
    def CopyDirLocal(srcPath, dstPath):
        s3File.CopyLocal(srcPath, dstPath, is_recursive=True)

    # @staticmethod
    # def CopyHdfs(srcPath, dstPat):
    #     cmd = 's3-dist-cp --src={srcPath} --dest={dstPath}'.format(srcPath=srcPath, dstPath=dstPath)
    #     if is_recursive:
    #         pass
    #     os.system(cmd)

    @staticmethod
    def CopyDirHdfs(srcPath, dstPath):
        cmd = 's3-dist-cp --src={srcPath} --dest={dstPath}'.format(srcPath=srcPath, dstPath=dstPath)
        os.system(cmd)

    @staticmethod
    def Copy(srcPath, dstPath, overwrite=False):
        if srcPath != dstPath:
            if overwrite and s3File.Exists(dstPath):
                s3File.Remove(dstPath)
            newbucket, newkey = s3File.split_s3_path(dstPath)
            oldname = srcPath[len("s3://"):]
            s3File.s3.Object(newbucket, newkey).copy_from(CopySource=oldname)

class pathStr(os.PathLike):
    def __new__(cls, *args, **kwargs):
        instance = super(pathStr, cls).__new__(cls)
        instance.path = None
        return instance

    def __init__(self, path, prefix=None):
        if isinstance(path, pathStr):
            self.path = path.path
            self.attr = path.attr
        else:
            if prefix is not None:
                self.path = os.path.join(prefix, path)
            else:
                self.path = path
            self._setAttr()
            if self.attr == 'local' and prefix is None:
                self.path = os.path.abspath(path)

    def _setAttr(self):
        if self.path.startswith('s3://'):
            self.attr = 's3'
        elif self.path.startswith('hdfs://'):
            self.attr = 'hdfs'
            if self.path.startswith('hdfs:///'):
                self.path = 'hdfs://'+hdfsFile.ip + self.path[7:]
        else:
            self.attr = 'local'

    def __add__(self, other):
        # self + other
        if other[0] == '/':
            other = other[1:]
        path = os.path.join(self.path, other)
        return pathStr(path)

    def __radd__(self, other):
        # self + other
        path = os.path.join(other, self.path)
        return pathStr(path)

    def __iadd__(self, other):
        if other[0] == '/':
            other = other[1:]
        self.path = os.path.join(self.path, other)
        return self

    def __getitem__(self, item):
        return self.path[item]

    def __fspath__(self):
        return self.path

    def __str__(self):
        return self.path

    def __repr__(self):
        return self.path

    def __getattr__(self, item):
        obj = getattr(self.path, item)
        if hasattr(obj, '__call__'):
            def wrapper(*args, **kwargs):
                result = obj(*args, **kwargs)
                if isinstance(result, str):
                    return pathStr(result)
                else:
                    return result
            return wrapper
        else:
            return obj

class s3FileObj(object):
    def __init__(self, fileobj):
        self.file = fileobj

    def read(self):
        return self.file.read().decode('utf-8')

    def readline(self):
        f = self.file.iter_lines()
        for line in f:
            yield line.decode('utf-8')

    def readlines(self):
        all_file = self.file.read().decode('utf-8')
        all_file_list = all_file.split('\n')
        return all_file_list

    def __enter__(self):
        pass

    def __exit__(self):
        self.file.close()


def addtail(path):
    if path[-1] != '/':
        path += '/'
    return path

def removetail(path):
    if path[-1] == '/':
        path = path[:-1]
    return path

if __name__ == '__main__':
    p = pathStr('/Users/snoworday')
    s3File.Walk(p)
    os.stat(p)
    print(str(p))
    p[2]
    p.__str__()
    p.strip()
    t = p + 'd'
    b = 4







import yaml

def writeFileToHDFS():
  rootdir = '/tmp/mnist_model'
  client = HdfsClient(hosts='localhost:50070')
  client.mkdirs('/user/root/mnist_model')
  for parent,dirnames,filenames in os.walk(rootdir):
    for dirname in  dirnames:
          print("parent is:{0}".format(parent))
    for filename in filenames:
          client.copy_from_local(os.path.join(parent,filename), os.path.join('/user/root/mnist_model',filename), overwrite=True)

def load_conf(conf_name):
    def wrapper(filename):
        with open('./{}/'.format(conf_name) + filename, 'r') as f:
            return yaml.load(f)
    return wrapper


import tensorflow as tf
import os
import numpy as np
from collections import OrderedDict, defaultdict
import yaml
from dl_rank.file import wfile
try:
    from dl_rank.layer import layers
except:
    from layer import layers

class BaseParser(object):
    def __init__(self, confPath, mode, useSpark):
        self.useSpark = useSpark
        self.confPath = confPath
        self.conf_dict = LoadDict(self.load_conf_switch(mode))
        self.column_to_csv_defaults()
        self.data_file_filter = lambda f_name: f_name.split('.')[-1] == 'txt'

    @property
    def model_out_format(self):
        if hasattr(self, '_model_out_format'):
            return self._model_out_format
        else:
            _model_out_format = self.conf_dict['separator']['model_out_format']
            self._model_out_format = [item+'_' for item in _model_out_format]
            return self._model_out_format

    def load_conf_switch(self, mode):
        confPath = self.confPath
        if self.useSpark:
            from pyspark import SparkFiles
            def wrapper(filename):
                try:
                    with open(SparkFiles.get(filename+'_'+mode+'.yaml'), 'r') as f:
                        return yaml.load(f)
                except:
                    with open(SparkFiles.get(filename+'.yaml'), 'r') as f:
                        return yaml.load(f)
            return wrapper
        else:
            def wrapper(filename):
                if wfile.Exists(os.path.join(confPath, filename+'_'+mode+'.yaml')):
                    with wfile.Open(os.path.join(confPath, filename+'_'+mode+'.yaml'), 'r') as f:
                        return yaml.load(f)
                else:
                    with wfile.Open(os.path.join(confPath, filename+'.yaml'), 'r') as f:
                        return yaml.load(f)
            return wrapper

    def load_all_conf(self, *confs):
        all_conf = dict()
        for conf in confs:
            all_conf.update({conf: self.conf_dict[conf]})
        return all_conf

    def column_to_csv_defaults(self):
        """parse columns to record_defaults param in tf.decode_csv func
        Return:
            OrderedDict {'feature name': [''],...}
        """
        feature_conf = self.conf_dict['feature']
        feature_list_conf = self.conf_dict['schema']
        feature_list = [feature_list_conf[key] for key in sorted(feature_list_conf, reverse=False)]
        feature_unused = []
        csv_defaults = OrderedDict()
        csv_scope = OrderedDict()
        for f in feature_list:
            if f in feature_conf and not feature_conf[f]['ignore']:  # used features
                conf = feature_conf[f]
                scope = feature_conf[f]['parameter']['scope'] if 'scope' in feature_conf[f]['parameter'] else 'embedding'
                name = feature_conf[f]['parameter']['name'] if 'name' in feature_conf[f]['parameter'] else f
                csv_scope[f] = [scope, name]
                if conf['type'] == 'category':
                    csv_defaults[f] = ['']
                else:
                    csv_defaults[f] = ['']  # 0.0 for float32
            else:  # unused feature
                feature_unused.append(f)
                csv_defaults[f] = ['']
        self.feature_unused = feature_unused
        self.column_defaults = csv_defaults
        self.column_scope = csv_scope

    def serving_parse_fn(self, pred_node_names):
        csv_defaults = self.column_defaults
        csv_scope = self.column_scope
        feature_unused = self.feature_unused
        model_out_format = self.model_out_format
        def serving_feature_receiver_fn():
            input_dict = dict()
            # scope_set = set([f[0] for f in csv_scope.values()])
            csv_scope_name = defaultdict(list)
            for feature, scope_name in csv_scope.items():
                csv_scope_name[scope_name[0]].append(feature)
            with tf.variable_scope('placeholder'):
                for scope, features in csv_scope_name.items():
                    with tf.variable_scope(scope):
                        for feature in features:
                            if feature in feature_unused:
                                continue
                            input_dict.update({feature: tf.compat.v1.placeholder(dtype=tf.string, shape=[None], name=feature)})
            input_dict.update({id: tf.compat.v1.placeholder(dtype=tf.string, shape=[None], name=id) for id in model_out_format if id[:-1] not in pred_node_names})
            return tf.estimator.export.ServingInputReceiver(features=input_dict, receiver_tensors=input_dict)
        return serving_feature_receiver_fn

    def parse_fn(self, data_type, isPred=False, na_value='', tail=''):
        csv_defaults = self.column_defaults
        feature_unused = self.feature_unused
        primary_delim = self.conf_dict['separator']['primary_delim']
        secondary_delim = self.conf_dict['separator']['secondary_delim']
        train_data_format = self.conf_dict['separator']['train_data_format']
        infer_data_format = self.conf_dict['separator']['infer_data_format']
        def _parser_feature(features):
            decode_features = tf.io.decode_csv(
                records=features, record_defaults=list(csv_defaults.values()),
                field_delim=secondary_delim, use_quote_delim=False, na_value=na_value)
            decode_features = dict(zip(csv_defaults.keys(), decode_features))
            for f in feature_unused:
                decode_features.pop(f)
            features_tail = {key+tail: decode_features[key] for key in decode_features}
            return features_tail

        def tfrecord_parser(value):
            if isPred:
                field_config = {field_name: tf.FixedLenFeature([], tf.string) for field_name in infer_data_format}
                batch_data = tf.parse_example(value, features=field_config)
                features = batch_data.pop("features")
                features_tail = _parser_feature(features)
                features_tail.update({elem+'_'+tail: data for elem, data in batch_data.items() if elem != 'features'})
                return features_tail
            else:
                field_config = {field_name: tf.FixedLenFeature([], tf.string) for field_name in train_data_format}
                batch_data = tf.parse_example(value, features=field_config)
                features = batch_data.pop("features")
                features_tail = _parser_feature(features)
                labels = tf.concat([tf.expand_dims(batch_data[label], 1) for label in train_data_format if label != 'features'], axis=1)
                labels = tf.strings.to_number(labels, out_type=tf.int32)
                return features_tail, labels

        def csv_parser(value):
            """Parse train and eval data with label
            Args:
                value: Tensor("arg0:0", shape=(), dtype=string)
            """
            if isPred:
                decode_data = tf.io.decode_csv(
                    records=value, record_defaults=[['']]*len(infer_data_format),
                    field_delim=primary_delim, use_quote_delim=False, na_value=na_value)
                data_container = dict(zip(infer_data_format, decode_data))
                features_tail = _parser_feature(data_container['features'])
                features_tail.update({elem+'_'+tail: data for elem, data in data_container.items() if elem != 'features'})
                return features_tail
            else:
                decode_data = tf.io.decode_csv(
                    records=value, record_defaults=[['']]*len(train_data_format),
                    field_delim=primary_delim, use_quote_delim=False, na_value=na_value)
                data_container = dict(zip(train_data_format, decode_data))
                features_tail = _parser_feature(data_container['features'])
                labels = [tf.equal(data_container[label], '1') for label in train_data_format if label != 'features']
                labels = tf.concat([tf.expand_dims(label, 1) if label.shape.ndims<2 else label for label in labels], axis=1)
                return features_tail, labels
        if data_type == 'tfrecord':
            return tfrecord_parser
        else:
            return csv_parser

    # def parse_fn(self, data_type, isPred=False, na_value='', tail=''):
    #     primary_delim = self.conf_dict['separator']['primary_delim']
    #     # secondary_delim = self.conf_dict['separator']['secondary_delim']
    #     train_data_format = self.conf_dict['separator']['train_data_format']
    #     infer_data_format = self.conf_dict['separator']['infer_data_format']
    #     feature_parse_fn = self.feature_parse_fn(data_type)
    #     def tfrecord_parser(value):
    #         if isPred:
    #             field_config = {field_name: tf.FixedLenFeature([], tf.string) for field_name in infer_data_format}
    #             batch_data = tf.parse_example(value, features=field_config)
    #             return batch_data["features"]
    #         else:
    #             field_config = {field_name: tf.FixedLenFeature([], tf.string) for field_name in train_data_format}
    #             batch_data = tf.parse_example(value, features=field_config)
    #             features = batch_data.pop("features")
    #             labels = tf.concat([tf.expand_dims(batch_data[label], 1) for label in train_data_format if label != 'features'], axis=1)
    #             labels = tf.strings.to_number(labels, out_type=tf.int32)
    #             return features, labels
    #
    #     def csv_parser(value):
    #         """Parse train and eval data with label
    #         Args:
    #             value: Tensor("arg0:0", shape=(), dtype=string)
    #         """
    #         if isPred:
    #             decode_data = tf.io.decode_csv(
    #                 records=value, record_defaults=[['']]*len(infer_data_format),
    #                 field_delim=primary_delim, use_quote_delim=False, na_value=na_value)
    #             data_container = dict(zip(infer_data_format, decode_data))
    #             features_tail = _parser_feature(data_container['features'])
    #             features_tail.update({elem+'_'+tail: data for elem, data in data_container.items() if elem != 'features'})
    #             return data_container['features']
    #         else:
    #             decode_data = tf.io.decode_csv(
    #                 records=value, record_defaults=[['']]*len(train_data_format),
    #                 field_delim=primary_delim, use_quote_delim=False, na_value=na_value)
    #             data_container = dict(zip(train_data_format, decode_data))
    #             labels = [tf.equal(data_container[label], '1') for label in train_data_format if label != 'features']
    #             labels = tf.concat([tf.expand_dims(label, 1) if label.shape.ndims<2 else label for label in labels], axis=1)
    #             return data_container['features'], labels
    #     if data_type == 'tfrecord':
    #         return tfrecord_parser
    #     else:
    #         return csv_parser

    def model_input_parse_fn(self, model_secondary_parse_fn):
        first_parser = self.first_parse_fn
        second_parser = self.secondary_parse_fn if hasattr(self, 'secondary_parse_fn') else model_secondary_parse_fn
        csv_defaults = self.column_defaults
        feature_unused = self.feature_unused
        secondary_delim = self.conf_dict['separator']['secondary_delim']
        teriary_delim = self.conf_dict['separator']['teriary_delim']
        def wrapper(features, params, dims, model_struct):
            sparse_emb, deep_emb, dense_emb, mask = first_parser(features, params, teriary_delim)
            features = second_parser(sparse_emb, deep_emb, dense_emb, mask, model_struct)
            return dims, features
        return wrapper

    @staticmethod
    def first_parse_fn(features, params, teriary_delim):
        '''

        :param features:
        :param params:
        :param is_input_indices: if True: input feature has been changed to index, else string
        :return:
        sparse_features: [batch, one_hot_cate_con]
        deep_features: [batch, [cate, con], embedding_size]
        dense_features: [batch, con_num]
        '''

        sparse = params['sparse']
        deep = params['deep']
        dense = params['dense']
        numeric = params['numeric']
        table = params['table']
        multi = params['multi']

        features = {key: tf.identity(features[key], name=key) for key in features}
        # batch_size = tf.shape(features[list(features.keys())[0]])[0]
        sparse_emb = OrderedDict()
        deep_emb = OrderedDict()
        dense_emb = OrderedDict()
        dense_emb_noreduce = OrderedDict()
        mask = dict()

        for f in features:
            f_type = multi[f][0]
            num = multi[f][1]
            fill_value = multi[f][6]
            features_f = tf.strings.split(features[f], sep=teriary_delim, maxsplit=num)
            if fill_value == '':
                fill_mask = tf.not_equal(features_f.values, '')
                len_mask = tf.less(features_f.indices[:, 1], abs(num))
                f_mask = tf.squeeze(tf.where(tf.logical_and(fill_mask, len_mask)), axis=1)
                features_f = tf.SparseTensor(indices=tf.gather(features_f.indices, f_mask),
                                             values=tf.gather(features_f.values, f_mask),
                                             dense_shape=(features_f.dense_shape[0], abs(num)))
            else:
                if f_type == 'category':
                    features_f_val = tf.regex_replace(features_f.values, '^$', fill_value)
                else:
                    mask_empty = tf.cast(tf.equal(features_f.values, ''), dtype=tf.float32)
                    features_f_val = tf.regex_replace(features_f.values, '^$', '0')
                    features_f_val = tf.strings.to_number(features_f_val) + fill_value * mask_empty
                f_mask = tf.squeeze(tf.where(tf.less(features_f.indices[:, 1], abs(num))), axis=1)
                features_f = tf.SparseTensor(indices=tf.gather(features_f.indices, f_mask),
                                             values=tf.gather(features_f_val, f_mask),
                                             dense_shape=(features_f.dense_shape[0], abs(num)))
            features[f] = features_f

        for f in numeric:
            f_type, num, size, combiner, same, default_value, _, null_value = multi[f]
            if combiner == 'none':
                features_f, f_mask = layers.to_dense(features[f], default_value=null_value, out_type=tf.float32)
                dense_emb_noreduce.update({f: (features_f, f_mask)})
                if f in dense:
                    dense_emb.update({f: features_f})    #chi cun shi fou wei N, 1 haishi N,
                if num < 0 and f not in mask.keys():
                    mask[f] = f_mask
            else:
                sparse_f_float, sparse_f_indice = layers.sparse_str2sparse_num(features[f], tf.float32)
                # tfprint = tf.print('sparse_f_float', sparse_f_float, summarize=1e6)
                # with tf.control_dependencies([tfprint]):
                f_input_combine = layers.sparse_reduce(sparse_f_float, reduce_type=combiner)# tf.cast(tf.expand_dims(tf.nn.embedding_lookup_sparse(tf.ones([abs(num)], dtype=tf.float32), sparse_f_indice, sparse_f_float, combiner='sqrtn'), 1), dtype=tf.float32)
                # tp = tf.print('sparse_f_indice', sparse_f_indice, 'sparse_f_float', sparse_f_float, 'combiner', f_input_combine, 'thereis f_input_combine', features[f], f, summarize=1e6)
                # with tf.control_dependencies([tp]):
                #     f_input_combine = tf.identity(f_input_combine)
                dense_emb_noreduce.update({f: (sparse_f_float if not same else f_input_combine, sparse_f_indice)})
                if f in dense:
                    dense_emb.update({f: f_input_combine})
        # all sparse only for category features
        for f in sparse:
            f_type, num, size, combiner, same, default_value, _, null_value = multi[f]
            if combiner == 'none':
                if f in table.keys():
                    features_f_indice = table[f].lookup(features[f])
                    features_f_indicator, f_mask = layers.to_dense(features_f_indice, default_value=int(null_value))
                    # f_mask = tf.not_equal(features_f_indicator, null_value)
                else:
                    features_f_indicator, f_mask = layers.to_dense(features[f], default_value=null_value, out_type=tf.int32)
                    features_f_indicator = layers.replace2default(features_f_indicator, size, default_value)
                # features_f_emb = tf.nn.embedding_lookup(sparse[f], features_f_indicator)
                features_f_emb = layers.multi_hot(features_f_indicator, num=abs(num), depth=size)
                sparse_emb.update({f: features_f_emb})
                if num < 0 and f not in mask.keys():
                    mask[f] = f_mask
            else:
                if f in table.keys():
                    features_f_indice = table[f].lookup(features[f])
                else:
                    features_f_indice, _ = layers.sparse_str2sparse_num(features[f], tf.int32)
                    features_f_indice = layers.replace2default(features_f_indice, size, default_value)
                # features_f_emb = tf.nn.embedding_lookup_sparse(sparse[f], features_f_indice, None, combiner=combiner)
                features_f_emb = layers.multi_hot(features_f_indice, depth=size, combiner=combiner)
                # _ = tf.reduce_mean(features_f_emb, axis=0, name='{}_embedding_params'.format(f))
                sparse_emb.update(
                    {f: tf.expand_dims(features_f_emb, 1)})

        for f in deep:
            f_type, num, size, combiner, same, default_value, _, null_value = multi[f]
            if f_type == 'category':
                if combiner == 'none':
                    # deep_f = tf.pad(deep[f], tf.constant([[0, 1], [0, 0]]), mode='CONSTANT',
                    #                 constant_values=0) if int(null_value) == size else deep[f]
                    deep_f = deep[f]
                    if f in table.keys():
                        features_f_indice = table[f].lookup(features[f])
                        features_f_indicator, f_mask = layers.to_dense(features_f_indice, default_value=null_value)
                        # f_mask = tf.not_equal(features_f_indicator, null_value)
                    else:
                        # tfprint2 = tf.print(f, features[f], tf.shape(features[f]), 'start', summarize=10000)
                        # with tf.control_dependencies([]):
                        features_f_indicator, f_mask = layers.to_dense(features[f], default_value=null_value, out_type=tf.int32)
                        features_f_indicator = layers.replace2default(features_f_indicator, size, default_value)
                    features_f_emb = layers.embedding_lookup(deep_f, features_f_indicator)
                    deep_emb.update({f: features_f_emb})
                    if num < 0:
                        mask[f] = f_mask
                        _ = tf.identity(tf.multiply(tf.cast(tf.expand_dims(f_mask, -1), tf.float32), features_f_emb), name='{}_used_embedding_params'.format(f))    #bs,num.vec
                    else:
                        _ = tf.identity(features_f_emb, name='{}_used_embedding_params'.format(f))#bs,num,vec
                else:
                    if f in table.keys():
                        features_f_indice = table[f].lookup(features[f])
                    else:
                        features_f_indice, _ = layers.sparse_str2sparse_num(features[f], tf.int32)
                        features_f_indice = layers.replace2default(features_f_indice, size, default_value)
                    if same:
                        features_f_emb = tf.tile(deep[f], tf.stack(tf.shape(features_f_indice.dense_shape)[0], 1))
                        assert False, 'u cant do this（set same in category feature） 0_0'
                    else:
                        features_f_emb = tf.nn.embedding_lookup_sparse(deep[f], features_f_indice, None, combiner=combiner)
                        features_f_emb = tf.pad(features_f_emb, [[0, tf.cast(features_f_indice.dense_shape[0], tf.int32)-tf.shape(features_f_emb)[0]], [0, 0]], 'CONSTANT', constant_values=0)
                        # tfprint3 = tf.print('features_f_emb_shape', features_f_indice.dense_shape, 'wait', tf.shape(features_f_emb), summarize=100000)
                        # with tf.control_dependencies([]):
                        features_f_emb = tf.identity(features_f_emb)
                    deep_emb.update(
                        {f: tf.expand_dims(features_f_emb, 1)})
                    _ = tf.identity(features_f_emb, name='{}_used_embedding_params'.format(f))
            elif f_type == 'numeric':
                features_f_value, sparse_f_indice = dense_emb_noreduce[f]
                if combiner == 'none':
                    features_f_emb = tf.multiply(tf.expand_dims(features_f_value, -1), tf.expand_dims(deep[f], 0))
                    deep_emb.update({f: features_f_emb})
                    sparse_f_indice = tf.cast(sparse_f_indice, tf.float32)
                    _ = tf.multiply(sparse_f_indice[:, :, np.newaxis], deep[f][np.newaxis, :, :], name='{}_used_embedding_params'.format(f))
                else:
                    if same:
                        features_f_combine = tf.expand_dims(tf.matmul(features_f_value, deep[f]), axis=1)
                        deep_emb.update({f: features_f_combine})
                        ###
                        _ = tf.identity(tf.expand_dims(deep[f], 0), name='{}_used_embedding_params'.format(f))#1,field_num,vec
                    else:
                        features_f_emb = layers.embedding_lookup_sparse(deep[f], sparse_f_indice, features_f_value, combiner=combiner)
                        _ = layers.embedding_lookup_sparse(deep[f], sparse_f_indice, None, combiner='mean', name='{}_used_embedding_params'.format(f))
                        # _ = tf.identity(features_f_emb, name='{}_used_embedding_params'.format(f))#bs,vec
                        features_f_combine = tf.expand_dims(features_f_emb, 1)
                        deep_emb.update({f: features_f_combine})

        return sparse_emb, deep_emb, dense_emb, mask

class LoadDict(dict):
    def __init__(self, load_fn):
        super(LoadDict, self).__init__()
        self.load_fn = load_fn

    def __getitem__(self, item):
        try:
            return dict.__getitem__(self, item)
        except KeyError:
            try:
                value = self.load_fn(item)
            except:
                value = None
            self.__setitem__(item, value)
            return value



import tensorflow as tf


def multi_hot(cat_int_tensor, depth, num=0, combiner=None):
    if isinstance(cat_int_tensor, tf.Tensor):
        out = tf.concat([tf.expand_dims(tf.one_hot(cat_int_tensor[:, n], depth, axis=-1), 1) for n in range(num)], axis=1)
    else:
        indices = tf.concat([tf.expand_dims(cat_int_tensor.indices[:, 0], axis=1), tf.expand_dims(cat_int_tensor.values, axis=1)], axis=1)
        updates = tf.ones_like(cat_int_tensor.values)
        shape = [cat_int_tensor.dense_shape[0], depth]
        if combiner == 'sum':
            out = tf.scatter_nd(indices, updates, shape)
        elif combiner == 'mean':
            row_indices = indices[:, 0]
            row_sum = tf.scatter_nd(tf.expand_dims(row_indices, axis=1), updates, [tf.shape(cat_int_tensor)[0]])
            updates = tf.math.divide(1, embedding_lookup(row_sum, row_indices))
            out = tf.scatter_nd(indices, updates, shape)
        elif combiner == 'max':
            out = tf.scatter_nd(indices, updates, shape)
            out = tf.math.minimum(out, 1)
        else:
            assert False, "category features'one-hot embedding only support sum/mean/max "
        out = tf.cast(out, dtype=tf.float32)
    return out

def sparse_reduce(sparse_tensor, reduce_type):
    bs = tf.cast(sparse_tensor.dense_shape[0], dtype=tf.int32)
    if reduce_type=='sum':
        sparse_tensor_reduce = tf.sparse.reduce_sum(sparse_tensor, axis=1, keepdims=True)
    elif reduce_type=='max':
        sparse_tensor_reduce = tf.sparse.reduce_max(sparse_tensor, axis=1, keepdims=True)
    elif reduce_type=='mean':
        sparse_tensor_reduce = tf.sparse.reduce_sum(sparse_tensor, axis=1, keepdims=True)
        indices = tf.cast(sparse_tensor.indices, tf.int32)
        line_number_indices = indices[:, 0]
        line_count = tf.expand_dims(tf.math.bincount(line_number_indices, minlength=bs, dtype=tf.float32), 1)
        sparse_tensor_reduce = tf.div_no_nan(sparse_tensor_reduce, line_count)
    else:
        assert False, "no this way"
    sparse_tensor_scatter = tf.reshape(sparse_tensor_reduce, shape=[bs, 1])
    return sparse_tensor_scatter

def embedding_lookup_sparse(embedding_params, sparse_indices, weights, combiner, name=None):
    bs = tf.cast(sparse_indices.dense_shape[0], dtype=tf.int32)
    if combiner != 'mean':
        embedding_tensor_reduce = tf.nn.embedding_lookup_sparse(embedding_params, sparse_indices, weights, combiner=combiner, name=name)
        embedding_tensor_scatter = tf.pad(embedding_tensor_reduce, [[0, bs-tf.shape(embedding_tensor_reduce)[0]], [0,0]], 'CONSTANT', constant_values=0)
    else:
        embedding_tensor_reduce = tf.nn.embedding_lookup_sparse(embedding_params, sparse_indices, weights, combiner='sum')
        embedding_tensor_scatter = tf.pad(embedding_tensor_reduce, [[0, bs-tf.shape(embedding_tensor_reduce)[0]], [0,0]], 'CONSTANT', constant_values=0)
        line_number = tf.cast(sparse_indices.indices[:, 0], tf.int32)
        line_count = tf.expand_dims(tf.math.bincount(line_number, minlength=bs, dtype=tf.float32), 1)
        embedding_tensor_scatter = tf.div_no_nan(embedding_tensor_scatter, line_count, name=name)
    return embedding_tensor_scatter

def sparse_cast(sparse_tensor, dtype):
    if sparse_tensor.dtype == dtype:
        return sparse_tensor
    elif sparse_tensor.dtype == tf.string:
        return tf.SparseTensor(indices=sparse_tensor.indices, dense_shape=sparse_tensor.dense_shape, values=tf.strings.to_number(sparse_tensor.values, out_type=dtype))
    elif dtype == tf.string:
        return tf.SparseTensor(indices=sparse_tensor.indices, dense_shape=sparse_tensor.dense_shape, values=tf.as_string(sparse_tensor.values))
    else:
        return tf.SparseTensor(indices=sparse_tensor.indices, dense_shape=sparse_tensor.dense_shape, values=tf.cast(sparse_tensor.values, dtype=dtype))

def replace2default(tensor, intercept, default_value):
    if isinstance(tensor, tf.Tensor):
        isEffect = tf.cast(tf.less(tensor, intercept), tf.int32)
        tensor = isEffect * tensor + (1-isEffect) * default_value
        return tensor
    else:
        values = tensor.values
        isEffect = tf.cast(tf.less(values, intercept), tf.int32)
        values = isEffect * values + (1-isEffect) * default_value
        return tf.SparseTensor(indices=tensor.indices, values=values, dense_shape=tensor.dense_shape)

def sparse_str2sparse_num(f_input, type):
    values = f_input.values
    indices = f_input.indices
    dense_shape = f_input.dense_shape
    values = tf.strings.to_number(values, out_type=type)
    f_out = tf.SparseTensor(indices=indices, values=values, dense_shape=dense_shape)
    sparse_f_indice = tf.SparseTensor(indices=indices,
                                      values=tf.squeeze(tf.split(indices, [1, 1], axis=1)[1], axis=1),
                                      dense_shape=dense_shape)
    return f_out, sparse_f_indice

def embedding_lookup(params, ids):
    ids_reshape = tf.reshape(ids, [tf.shape(ids)[0]*tf.shape(ids)[1], 1])
    ids_reshape_sparse = to_sparse(ids_reshape, [tf.shape(ids_reshape)[0], 1])
    features_f_emb_reshape = tf.nn.embedding_lookup_sparse(params, ids_reshape_sparse, None, combiner="sum")
    features_f_emb = tf.reshape(features_f_emb_reshape, [tf.shape(ids)[0], ids.shape.as_list()[1], features_f_emb_reshape.shape.as_list()[-1]])
    return features_f_emb

def to_sparse(dense_tensor, dense_shape):
    indices = tf.where(tf.not_equal(dense_tensor, -1))
    values = tf.gather_nd(dense_tensor, indices)
    sparse = tf.SparseTensor(indices, values, dense_shape)
    return sparse

def trunk_sparse_tensor(sparse_tensor, length):
    indice = sparse_tensor.indices
    values = sparse_tensor.values
    index = tf.squeeze(tf.where(tf.less(indice[:, 1], length)), axis=1)
    return tf.SparseTensor(indice=tf.gather(indice, index),
                           values=tf.gather(values, index),
                           dense_shape=sparse_tensor.dense_shape)

def to_dense(tensor, default_value=0, out_type=None):
    # bs = tf.cast(tensor.dense_shape[0], tf.int32)
    if out_type is not None:
        tensor = sparse_cast(tensor, dtype=out_type)
    tensor_dense = tf.sparse.to_dense(tensor, default_value=default_value)
    # tfprint = tf.print('dense_shape', tf.shape(tensor_dense), 'tensor_dense', tensor_dense,  'tensor', tensor, 'tensor shape', tf.shape(tensor), summarize=100000)
    # with tf.control_dependencies([]):
    tensor_sparse_mask = tf.SparseTensor(indices=tensor.indices, values=tf.ones_like(tensor.values), dense_shape=tensor.dense_shape)
    tensor_mask = tf.sparse.to_dense(tensor_sparse_mask, default_value=0)
    # tensor_pad = tf.pad(tensor_dense, [[0, bs-tf.shape(tensor_dense)[0]], [0,depth-tf.shape(tensor_dense)[1]]], 'CONSTANT', constant_values=default_value)
    # tensor_reshape = tf.reshape(tensor_pad, [-1, depth])
    return tensor_dense, tensor_mask

def _init_weights_bias(input_dim, out_dim):
    weights = tf.Variable(
        tf.random.normal([input_dim, out_dim], 0.0, 0.01)
    )
    bias = tf.Variable(
        tf.random.normal([out_dim], 0.0, 0.01)
    )
    return weights, bias

def dense(input, out_dim, activation=None, bn=False, training=True):
    # w_out, b_out = self._init_weights_bias(int(input.shape[1]), out_dim)
    # output = tf.compat.v1.nn.dense(input, w_out, b_out)
    output = tf.layers.dense(input, out_dim, trainable=training)
    if bn:
        output = tf.compat.v1.layers.batch_normalization(output, training=training)
    if activation is not None:
        output = activation(output)
    return output

def focal_loss_sigmoid(labels,logits,weights=1,alpha=0.25,gamma=2):
    """
    Computer focal loss for binary classification
    Args:
      labels: A int32 tensor of shape [batch_size].
      logits: A float32 tensor of shape [batch_size].
      alpha: A scalar for focal loss alpha hyper-parameter. If positive samples number
      > negtive samples number, alpha < 0.5 and vice versa.
      gamma: A scalar for focal loss gamma hyper-parameter.
    Returns:
      A tensor of the same shape as `lables`
    """
    y_pred=tf.nn.sigmoid(logits)
    labels=tf.to_float(labels)
    L=-labels*(1-alpha)*((1-y_pred)*gamma)*tf.log(y_pred)-\
      (1-labels)*alpha*(y_pred**gamma)*tf.log(1-y_pred)
    ratio = (tf.cast(tf.shape(labels)[0], dtype=tf.float32)) / (tf.reduce_sum(weights) + 1)
    L = L * weights
    L = tf.reduce_mean(L) * ratio
    return L

def l1_loss(l):
    pass

def l2_loss(l):
    pass

def MBA_loss(l):
    nodes = {k: v for k, v in tf.get_default_graph()._nodes_by_name.items() if 'used_embedding_params' in k}
    mini_batch_aware_reg = 0
    for _, node in nodes.items():
        try:
            mini_batch_aware_reg += tf.reduce_sum(tf.reduce_mean(tf.square(node.outputs), axis=0))
        except:
            a=1
    return mini_batch_aware_reg * l



import tensorflow as tf
from .BaseModel import baseModel
try:
    from dl_rank.layer import layers
except:
    from layer import layers

class xdeepfm(baseModel):
    name = 'xdeepfm'
    def forward(self, wide_features, deep_features, dense_features, deep_embedding_columns, load_conf, is_training):
        with tf.variable_scope('xdeepfm'):
            deep_input = tf.split(deep_features, num_or_size_splits=self.model_conf['field_num'], axis=1)
            deep_input = tf.concat([tf.expand_dims(feature, axis=1) for feature in deep_input], axis=1)
            cross_layer_size = self.model_conf['cross_layer_size']
            final_result = []
            final_len = 0
            embed_dim = self.model_conf['embed_dim']
            field_nums = [self.model_conf['field_num']]
            cin_layers = [deep_input]
            for idx, layer_size in enumerate(cross_layer_size):
                dot_result = tf.einsum('ipk,iqk->ipqk', cin_layers[0], cin_layers[-1])
                dot_result = tf.reshape(dot_result, shape=[-1, embed_dim, field_nums[0] * field_nums[-1]])
                filters = tf.get_variable(initializer=self.initializer_fn, name='f_' + str(idx),
                                          shape=[1, field_nums[-1] * field_nums[0], layer_size],
                                          dtype=tf.float32)

                # [batch_size, embed_dim, layers_size]
                curr_out = tf.nn.conv1d(dot_result, filters=filters, stride=1, padding='VALID')
                final_len += layer_size
                field_nums.append(layer_size)
                final_result.append(curr_out)
                curr_out = tf.transpose(curr_out, [0, 2, 1])
                cin_layers.append(curr_out)
            result = tf.concat(final_result, axis=2)
            result = tf.reduce_sum(result, axis=1)  # [batch_size, layer_size]
            y_xdeepfm = tf.layers.dense(result, 1, activation=None,
                                        kernel_initializer=self.initializer_fn)
        with tf.variable_scope('dnn'):
            dnn_hidden_units = self.model_conf['dnn_hidden_units']
            deep_input = tf.concat([deep_features, dense_features], axis=1)
            for dim in dnn_hidden_units:
                deep_input = tf.layers.dense(deep_input, dim, activation=self.activation_fn,
                                             kernel_initializer=self.initializer_fn)
            y_dnn = tf.layers.dense(deep_input, 1, activation=None,
                                    kernel_initializer=self.initializer_fn)

        return y_xdeepfm + y_dnn + tf.reduce_sum(wide_features, axis=1, keep_dims=True)